'''Utility functions.'''
import math
import tkinter as tk


TK_VERSION = tuple(int(n) for n in str(tk.TkVersion).split('.'))
'''
Get the tk version as an integer tuple.

Similar to `sys.version_info`.
'''


def lcm_multiple(*numbers):
    '''
    Least Common Multiple: Multiple number

    .. note::

        Python 3.9 has `math.lcm <https://docs.python.org/3.9/library/math.html?highlight=math%20lcm#math.lcm>`_.
    '''
    if len(numbers) > 0:
        lcm = numbers[0]
        for n in numbers[1:]:
            lcm = lcm_single(lcm, n)
        return lcm
    else:
        return None


def lcm_single(a, b):
    '''
    Least Common Multiple: Single Pair
    '''
    if a == 0 and b == 0:
        return 0
    else:
        return int((a * b) / math.gcd(a, b))



# Configuration file for the Sphinx documentation builder.
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Imports -----------------------------------------------------------------
import os
from datetime import datetime
from importlib.metadata import version as metadata_version, metadata
from pathlib import Path
import json

from sphinx.util import logging
logger = logging.getLogger(__name__)

needs_sphinx = '6.1'

# -- Automatic Information ---------------------------------------------------
now = datetime.now()
year = '%d' % now.year

# `production`/`debug` tags
# - Can be used on the document as:
#   .. only:: production
#   .. only:: debug
# - Defaults to debug mode
PRODUCTION = tags.has('production')  # noqa: F821
if not PRODUCTION:
    tags.add('debug')  # noqa: F821

# Tk Version
try:
    import tkinter as tk
    tk_version = tk._tkinter.TK_VERSION
except ImportError:
    tk_version = ''

# -- Project information -----------------------------------------------------
project = os.environ.get('READTHEDOCS_PROJECT', os.environ['PROJECT'])
metadata_project = metadata(project)
author = metadata_project.get('Author', 'Powertools Technologies')
copyright = f'{year}, {author}'

category = 'Miscellaneous'
description = metadata_project.get('Summary', f'{project} Description')

version = metadata_version(project)
release = version if PRODUCTION else 'dev'

# Primary Project Domain
# https://www.sphinx-doc.org/en/master/usage/restructuredtext/domains.html
primary_domain = 'py'

# The language for content autogenerated by Sphinx, and for marking the docs.
# https://www.sphinx-doc.org/en/master/usage/configuration.html#confval-language
# See Also: https://www.sphinx-doc.org/en/master/usage/advanced/intl.html#intl
# language = None # This is given in `SPHINX_LANGUAGE` on the `Makefile`

# -- Project Settings --------------------------------------------------------

# The suffix(es) of source filenames.
source_suffix = {
    '.rst': 'restructuredtext',
    # '.md': 'markdown',
}

# The master toctree document.
root_doc = 'index'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    'Thumbs.db',
    '.DS_Store',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# Show warnings on the built docs. Useful for debugging
keep_warnings = PRODUCTION is False
# List of warnings to suppress
suppress_warnings = [
]


# Default role
# `txt` is equivalent to :DEFAULT_ROLE:``
default_role = 'any'

today_fmt = '%d %B %Y'

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# Default highlight language
highlight_language = 'none'

# Automatically number figures (with caption)
numfig = True

# -- Sphinx Extensions ----------------------------------------------------

extensions = []

# InterSphinx
# Link to other Sphinx documentation repositories
extensions.append('sphinx.ext.intersphinx')
# - See `release/docs-intersphinx.py` for this setup
intersphinx_cache_folder = Path('_inventories')
intersphinx_cache_metadata_fname = intersphinx_cache_folder / 'metadata.json'
if intersphinx_cache_metadata_fname.is_file():
    logger.info('load intersphinx metadata from %s...', intersphinx_cache_metadata_fname)
    with intersphinx_cache_metadata_fname.open('r') as fobj:
        intersphinx_cache_metadata = json.load(fobj)
else:
    intersphinx_cache_metadata = {
    }
# Defined in `release/docs-intersphinx.py`, on the "MAPPING" variable
intersphinx_mapping = {
    key: (
        url,
        (str(intersphinx_cache_folder / f'{key}.inv'), None),
    )
    for key, url in intersphinx_cache_metadata.items()
}
intersphinx_cache_limit = -1  # Forever
intersphinx_timeout = 5  # seconds

# MathJAX - https://docs.mathjax.org/en/latest/installation.html#installation
# By default uses the CDN AKA an external server.
# See the installation URL for a local install, but it's big (>30MB)
#  extension.append('sphinx.ext.mathjax')

# GraphViz
# Use SVG instead of rendering to PNG
extensions.append('sphinx.ext.graphviz')
graphviz_output_format = 'svg'

# Automatically Label Sections
# Add a reference to all document sections.
# https://www.sphinx-doc.org/en/master/usage/extensions/autosectionlabel.html
extensions.append('sphinx.ext.autosectionlabel')
autosectionlabel_prefix_document = True  # Prefix reference with document name

# AutoDoc
extensions.append('sphinx.ext.autodoc')
autodoc_member_order = 'groupwise'
autodoc_default_options = {
    'members': True,
    'show-inheritance': True,
}
autodoc_typehints = 'both'
autodoc_typehints_description_target = 'documented'

# AutoDoc: Napoleon
# Convert Google-style docstring to proper rst metadata
extensions.append('sphinx.ext.napoleon')
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_use_admonition_for_examples = True
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = True
napoleon_attr_annotations = True

# AutoSummary
extensions.append('sphinx.ext.autosummary')
autosummary_generate = True
autosummary_generate_overwrite = False

# Auto API Documentation
# - A souped-up version of AutoSummary and AutoDoc
extensions.append('sphinx_automodapi.automodapi')
extensions.append('sphinx_automodapi.smart_resolver')

# Expose slow files
# - Only in production
if PRODUCTION:
    extensions.append('sphinx.ext.duration')

# Support ToDo
extensions.append('sphinx.ext.todo')
todo_emit_warnings = PRODUCTION is False
todo_include_todos = PRODUCTION is False

# Shorten External Links
extensions.append('sphinx.ext.extlinks')
extlinks = {
    # lpformat: Local Project in another Format, current location.
    #           Technically not an "external" link.
    #           The argument is the file extension. Usually only "pdf" is available.
    #           Usage:
    #             :lformat:`here <pdf>`
    #             :lformat:`pdf`
    'lpformat': (f'{project}.%s', 'Format[%s]'),
    # Tcl: TclCmd online manpages, for the correct version
    'tcl': (f'https://tcl.tk/man/tcl{tk_version}/TclCmd/%s', '``Tcl`` %s'),
    # TclTk: TkCmd online manpages, for the correct version
    'tk': (f'https://tcl.tk/man/tcl{tk_version}/TkCmd/%s', '``Tk`` %s'),
    # TclTk Library: TkLib online manpages, for the correct version
    'tk_lib': (f'https://tcl.tk/man/tcl{tk_version}/TkLib/%s', '``Tk`` %s'),
    # effbot: An Introduction to Tkinter by Fredrik Lundh
    'tkinter_effbot': ('https://effbot.org/tkinterbook/%s', '``effbot`` %s'),
    # NMT Computer Center tkinter8.5 reference guide
    # - Original URL: http://www.nmt.edu/tcc/help/pubs/tkinter/web/index.html
    # - Wayback Machine: https://web.archive.org/web/20171129043813/http://infohost.nmt.edu/tcc/help/pubs/tkinter/web/%s
    #   - Do not use, no images
    'tkinter_nmt': ('https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/%s', 'NMT %s'),
}
extlinks_detect_hardcoded_links = True  # Check hardcoded links that can use `extlinks`

# See Also:
# # extensions.append('sphinx.ext.ifconfig')
# # extensions.append('sphinx.ext.doctest')
# https://github.com/wpilibsuite/sphinxext-opengraph
# https://github.com/executablebooks/sphinx-autobuild#readme
# https://github.com/executablebooks/sphinx-copybutton#readme

# Optional:
# Include GitHub Pages Detritus
extensions.append('sphinx.ext.githubpages')

# -- Options for HTML output -------------------------------------------------
html_theme = 'furo'

# Theme: Alabaster (default)
# https://alabaster.readthedocs.io/en/latest/customization.html#theme-options
if html_theme == 'alabaster':
    pass

# Theme: Read the Docs
# https://sphinx-rtd-theme.readthedocs.io/en/stable/configuring.html#theme-options
if html_theme == 'sphinx_rtd_theme':
    html_theme_options = {
        'collapse_navigation': False,
        'navigation_depth': '-1',
        'display_version': True,
        'logo_only': False,
        'prev_next_buttons_location': None,
        'style_external_links': True,
        # 'style_nav_header_background': None,
    }

# Theme: furo
# https://pradyunsg.me/furo/
if html_theme == 'furo':
    html_theme_options = {
        'light_css_variables': {
            'color-brand-primary': '#FF6600',
            'color-brand-content': '#FF6600',
        },
        'dark_css_variables': {
            'color-brand-primary': '#FF6600',
            'color-brand-content': '#FF6600',
        },
    }
    navigation_with_keys = True
    pygments_dark_style = 'monokai'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Don't include the source files in the HTML output
html_copy_source = False

# Use this favicon for the HTML documents.
# Relative to the documentation folder, `../` is the repository root.
# https://www.sphinx-doc.org/en/master/usage/configuration.html#confval-html_favicon
html_favicon = '../release/icon.ico'

# Base URL locations. To be filled at runtime, based on `$BASEURL`
html_baseurl = os.environ.get('BASEURL', '')
html_use_opensearch = html_baseurl

# https://www.sphinx-doc.org/en/master/usage/configuration.html#confval-manpages_url
manpages_url = 'https://man.archlinux.org/man/{page}.{section}'

# -- Options for LaTeX output ---------------------------------------------

# One entry per LaTeX file. List of tuple:
# (source start file, target name, title, author, documentclass [howto, manual, or own class], toctree_only)
latex_documents = [
    # "Main" document, section 7: Miscellaneous
    (root_doc, '%s.tex' % project, '%s' % project, author, 'manual', False),
]

latex_engine = 'xelatex'
latex_logo = '_images/logo_horizontal.png'  # Show on the top of the title page. `None` to disable.
latex_toplevel_sectioning = 'part'
latex_domain_indices = True  # Generate domain indexes
latex_show_pagerefs = True  # Show page references on PDF
latex_show_urls = 'footnote'  # Show URL as footnotes
latex_use_xindy = True  # Use `xindy` instead of `makeindex`
latex_additional_files = [
    'sphinx.sty',  # Default Sphinx Package
    'custom.sty',
]

latex_elements = {
    'papersize': 'a4paper',  # A4 paper
    'pointsize': '12pt',  # Body font size
    'passoptionstopackages': r'\PassOptionsToPackage{svgnames}{xcolor}',
    'fncychap': r'',  # No default fancy headers
    'releasename': 'Release',
    'printindex': '',  # No final index
    'preamble': r'''
    \usepackage{custom}
    ''',
}

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    # "Main" manual page, section 7: Miscellaneous
    (root_doc, project, '%s Documentation' % project, [author], 7)
]
manpage_folder = Path('manpage')
if manpage_folder.is_dir():
    for fpath in manpage_folder.glob('*.rst'):
        fbase = fpath.stem
        man_description = None
        fdesc = fpath.with_suffix('.desc')
        if fdesc.is_file():
            with fdesc.open('r') as f:
                man_description = next(f)  # Use only the first line
        if not isinstance(man_description, str):  # Fallback description
            man_description = '%s Manual' % fbase
        logger.info('[standalone manpage] create "%s"', fbase)
        man_pages.append((str(manpage_folder / fbase), str(fbase), man_description, [author], 1))

# Add URL addresses after links
man_show_urls = True

# Create the MANPATH structure
man_make_section_directory = True

# Allow for `{}` to control emphasis. Example: `.. option:: -c <{FILE}>`
# https://www.sphinx-doc.org/en/master/usage/configuration.html#confval-option_emphasise_placeholders
option_emphasise_placeholders = True

# -- Options for Texinfo output -------------------------------------------

# One entry per texinfo document. List of tuples
# (source start file, target name, title, author, dir menu entry, description, category, toctree_only)
texinfo_documents = [
    # "Main" texinfo page, category: Miscellaneous
    (root_doc, project, '%s Documentation' % project, '@*'.join([author]), project, description, 'Miscellaneous', False),
]

texinfo_show_urls = 'footnote'  # Show URL as footnotes
texinfo_no_detailmenu = False  # Generate detailed menu on "Top"
texinfo_cross_references = True  # Generate cross references

# -- Options for linkcheck ------------------------------------------------

# Regexes with ignored URL
linkcheck_ignore = [
    # NMT Computer Center tkinter8.5 reference guide
    # - This page is now offline
    'http://www.nmt.edu/tcc/help/pubs/tkinter/web/index.html',
]

# Map URL regex to headers to be sent
linkcheck_request_headers = {
    "*": {
        "Accept": "text/html,application/xhtml+xml",
    }
}

# Check `#anchors` in URL, by parsing the result
linkcheck_anchors = True

linkcheck_retries = 5



# import typing
import unittest
import logging

from tkmilan import autolayout
from tkmilan.model import GridCoordinates as GC

logger = logging.getLogger(__name__)


class Test_autolayout(unittest.TestCase):
    SINGLE = [
        autolayout.AUTO,
        autolayout.HORIZONTAL,
        autolayout.VERTICAL,
        'X', 'xN', 'xS',
        'R1', 'r1', 'H1',
        'C1', 'c1', 'V1',
    ]

    MULTIPLE = [
        ('xE', [  # Square, grow right
            GC(0, 0), GC(0, 1),
            GC(1, 0), GC(1, 1),
        ]),
        ('xW', [  # Square, grow left
            GC(0, 1), GC(0, 0),
            GC(1, 1), GC(1, 0),
        ]),
        ('xS', [  # Square minus 1, grow down
            GC(0, 0), GC(1, 0),
            GC(0, 1, rowspan=2),
        ]),
        ('xH', [  # Square, grow right, reversed
            GC(1, 0), GC(1, 1),
            GC(0, 0), GC(0, 1),
        ]),
        ('R2,1', [  # Multiple, grow right
            GC(0, 0), GC(0, 1),
            GC(1, 0, columnspan=2),
        ]),
        ('H2,1', [  # Multiple, grow right, reversed
            GC(1, 0), GC(1, 1),
            GC(0, 0, columnspan=2),
        ]),
        ('C2,1', [  # Multiple, grow bottom
            GC(0, 0), GC(1, 0),
            GC(0, 1, rowspan=2),
        ]),
        ('V2,1', [  # Multiple, grow bottom, reversed
            GC(0, 1), GC(1, 1),
            GC(0, 0, rowspan=2),
        ]),
    ]

    def setUp(self):
        # "Beautify" the GridCoordinates `repr`
        GC.__repr__ = GC.__str__

    def test_singlewidget(self):
        logger.debug('Testing single widgets...')
        for layout in self.SINGLE:
            with self.subTest(layout=layout):
                self.assertEqual(list(autolayout.do(layout, 1)[1]), [
                    GC(0, 0),
                ])

    def test_multiplewidgets(self):
        logger.debug('Testing multiple widgets...')
        for layout, result in self.MULTIPLE:
            amount = len(result)
            with self.subTest(layout=layout, amount=amount):
                self.assertEqual(list(autolayout.do(layout, amount)[1]), result)


if __name__ == '__main__':
    import sys
    logs_lvl = logging.DEBUG if '-v' in sys.argv else logging.INFO
    logging.basicConfig(level=logs_lvl, format='%(levelname)5.5s:%(funcName)s: %(message)s', stream=sys.stderr)
    unittest.main()



'''Renderer for `diagrams <diagram.Diagram>`.

Can support multiple backends, only one implemented so far.
'''
import typing
import logging
import numbers
from functools import singledispatchmethod
import tkinter as tk

from . import diagram
from . import exception
from . import util
if typing.TYPE_CHECKING:
    from . import Canvas

logger = logging.getLogger(__name__)


TKCANVAS_COLORMAP_TEXT = {
    'fill': 'fill',
}
TKCANVAS_COLORMAP_MULTILINE = {
    'outline': 'fill',
}


def tkcanvas_width(**options: typing.Optional[int]) -> typing.Mapping:
    '''Generate ``Tk`` canvas width arguments'''
    rval = {}
    for oname, option in options.items():
        if option is not None:
            rval[oname] = float(option)
    return rval


def tkcanvas_dash(**options: typing.Optional[diagram.D]) -> typing.Mapping:
    '''Generate related `diagram.D` ``Tk`` canvas dash arguments'''
    rval = {}
    for oname, option in options.items():
        if option is not None:
            rval[oname] = option.pattern
            if oname == 'dash':
                rval['dashoffset'] = option.offset  # type: ignore[assignment]
            # TODO: Implement missing offsets?
    return rval


def tkcanvas_arrow(option: typing.Optional[diagram.A]) -> typing.Mapping:
    '''Generate related `diagram.A` ``Tk`` canvas arrow arguments'''
    if option is None:
        return {}
    else:
        assert (option.atStart, option.atEnd) != (False, False)
        return {
            'arrow': {
                (True, False): tk.FIRST,
                (False, True): tk.LAST,
                (True, True): tk.BOTH,
            }[option.atStart, option.atEnd],
            'arrowshape': (option.d1, option.d2, option.d3),
        }


def tkcanvas_smooth(option: typing.Optional[diagram.Smooth]) -> typing.Mapping:
    '''Generate related `diagram.Smooth` ``Tk`` canvas smooth arguments'''
    if option is None:
        return {}
    else:
        return {
            'smooth': True,  # TODO: Implement Bézier selection? Use `cnf`
            'splinesteps': option.steps,
        }


def tkcanvas_colors(noprefix: diagram.C,
                    cnames: typing.Optional[typing.Mapping[str, str]] = None,
                    **poptions: diagram.C) -> typing.Mapping:
    '''Generate related `diagram.C` ``Tk`` canvas color arguments.

    Args:
        noprefix: The color without prefix. Required.
        poptions: Mapping between prefix and color oboject.
        cnames: A mapping between color object option name, and ``Tk`` argument.
            Optional, defaults to ``fill`` and ``outline`` mapping to the
            corresponding names.
    '''
    assert '' not in poptions
    poptions[''] = noprefix
    rval = {}
    for oprefix, option in poptions.items():
        for cname, oname in (cnames or {s: s for s in ('fill', 'outline')}).items():
            cvalue = getattr(option, cname, None)
            if cvalue is not None:
                rval[f'{oprefix}{oname}'] = cvalue
    return rval


def tkcanvas_text_angle(angle: numbers.Real) -> typing.Mapping:
    '''Generate ``Tk`` `Text <diagram.Text>` ``angle`` arguments.

    This feature is only supported on later Tk versions.

    Args:
        angle: The angle value. Required.
    '''
    rval = {}
    if util.TK_VERSION >= (8, 6):
        rval['angle'] = str(angle)
    else:
        assert angle == 0.0, f'Unsupported Text.angle @ Tk {util.TK_VERSION}'
    return rval


# Renderer Implementation
class Renderer_TkCanvas:
    '''Renderer object for the ``Tk`` canvas backend.

    This is just a holding for the rendering process. No "automatic" draws
    happens when creating this object, all must go through the `redraw`
    function.

    Args:
        canvas: The ``Tk`` canvas object to use.
            Should be an internal `Canvas`.
        diagram: The diagram to render.
    '''
    def __init__(self, canvas: 'Canvas', diagram: diagram.Diagram):
        self._previous_size: typing.Tuple[int, int] = (-1, -1)
        self._e_clickelement: typing.Optional[str] = None
        self.diagram = diagram
        self.cv = canvas

    def redraw(self, _a1: typing.Any = None, _a2: typing.Any = None, *,
               width: typing.Optional[int] = None,
               height: typing.Optional[int] = None,
               force: bool = False,
               ):
        '''Redraw the canvas.

        The canvas consists in three layers, as explained in `diagram.Diagram`.

        If the canvas size changed, all layers are redrawn. Othewise, only the
        foreground is redrawn. Note that redrawing implies removing the old
        elements and creating new ones, this is done automatically.

        When the foreground is redrawn, it is pushed behing the background
        front layer.

        Unless ``force`` is given, the first "weird" event (when width and
        height are both ``1``) is skipped, since this does not make sense to
        redraw.

        Args:
            width: Use this canvas width, instead of getting this information
                from the ``canvas`` object.
            height: Use this canvas height, instead of getting this information
                from the ``canvas`` object.
            force: Force redrawing of background and foreground elements.

            _a1: Optional, unused. This exists for API compatibility with
                bindings.
            _a2: Optional, unused. This exists for API compatibility with
                traces.
        '''
        cv = self.cv
        cwidth = width or cv.winfo_width()
        cheight = height or cv.winfo_height()
        current_size = (cwidth, cheight)
        if not force and current_size == (1, 1):
            # The first weird Configure
            assert len(cv.find_all()) == 0
            if __debug__:
                logger.debug('Skip first weird `Configure` event: Size %s', current_size)
            return
        logger.debug('| Size: %s', current_size)
        to_skip = False
        minwidth, minheight = self.diagram.MIN_SIZE
        if not force:
            if minwidth is not None:
                if __debug__:
                    if cwidth < minwidth:
                        logger.debug('- Min Width: %d', minwidth)
                to_skip = to_skip or (cwidth < minwidth)
            if minheight is not None:
                if __debug__:
                    if cheight < minheight:
                        logger.debug('- Min Height: %d', minheight)
                to_skip = to_skip or (cheight < minheight)
        redraw_bg: bool = force or (current_size != self._previous_size)
        redraw_fg: bool = force or True  # TODO: Implement state tracker
        if to_skip:
            redraw_bg = redraw_fg = False
        if to_skip or (redraw_bg and redraw_fg):
            if __debug__:
                logger.debug('- Clean BG+FG (%d+%d)', len(cv.find_withtag(':bg')), len(cv.find_withtag(':fg')))
            assert set(cv.find_withtag(':bg || :fg')) == set(cv.find_all())
            cv.delete(tk.ALL)
        elif redraw_fg:
            if __debug__:
                logger.debug('- Clean FG (%d)', len(cv.find_withtag(':fg')))
            assert redraw_bg is False
            cv.delete(':fg')
        else:
            # Skip redraw
            assert redraw_bg is False and redraw_fg is False
        drew_fg = False
        if __debug__:
            drew_bg_b, drew_bg_f = False, False
        # TODO: Add `:latest` tag everywhere
        if redraw_bg:
            if __debug__:
                logger.debug('- Redraw BG:Back')
            tags_bg = [':bg', ':bg:b']
            for eraw in self.diagram.setup_bg_b(cwidth=cwidth, cheight=cheight):
                for element in eraw.iterate():
                    eid = self.render(element, extratags=tags_bg)
                    if eid is None:
                        if __debug__:
                            logger.warning('  - %s', element)
                        else:
                            raise exception.InvalidRender(element)
                    else:
                        logger.debug('  - %d: %s', eid, element)
                        if __debug__:
                            drew_bg_b = True
        if redraw_fg:
            if __debug__:
                logger.debug('- Redraw FG')
            tags_fg = [':fg']
            for eraw in self.diagram.setup_fg(cwidth=cwidth, cheight=cheight):
                for element in eraw.iterate():
                    eid = self.render(element, extratags=tags_fg)
                    if eid is None:
                        if __debug__:
                            logger.warning('  - %s', element)
                        else:
                            raise exception.InvalidRender(element)
                    else:
                        logger.debug('  - %d: %s', eid, element)
                        drew_fg = True
        if redraw_bg:
            if __debug__:
                logger.debug('- Redraw BG:Front')
            tags_bg = [':bg', ':bg:f']
            for eraw in self.diagram.setup_bg_f(cwidth=cwidth, cheight=cheight):
                for element in eraw.iterate():
                    eid = self.render(element, extratags=tags_bg)
                    if eid is None:
                        if __debug__:
                            logger.warning('  - %s', element)
                        else:
                            raise exception.InvalidRender(element)
                    else:
                        logger.debug('  - %d: %s', eid, element)
                        if __debug__:
                            drew_bg_f = True
        if drew_fg and len(cv.find_withtag(':bg:f')):
            if __debug__:
                logger.debug('- Lower "FG" below "BG:Front"')
            cv.tag_lower(':fg', ':bg:f')
        if __debug__:
            drew_map = {
                'BG:Back': drew_bg_b,
                'FG': drew_fg,
                'BG:Front': drew_bg_f,
            }
            drew_list = [lbl for lbl, b in drew_map.items() if b]
            logger.debug('Drew %d elements: %s', len(drew_list), ' '.join(drew_list))
            logger.debug('| Current Elements: %d', len(cv.find_all()))
        # TODO. Remove `:latest` tag, after logging the count
        # Setup Bindings
        e_clickelement = self.cv.tag_bind(tk.ALL, '<Button-1>', self.cv.onClickElement,
                                          add=False)  # Replace existing binding
        # Store information
        self._previous_size = current_size
        self._e_clickelement = e_clickelement

    @singledispatchmethod
    def render(self, de: diagram.DiagramElement, *, extratags: typing.Sequence[str] = tuple()) -> int:
        '''Render a diagram element, with some internal ``extratags``.

        This function is defined as a `single dispatch method
        <functools.singledispatchmethod>` for each supported diagram element.

        Args:
            de: The diagram element to render.
                Supported for the given diagram element types.
            extratags: Additional internal tags to consider, besides the
                element tags.
        '''
        # Fallback method for missing classes
        cls = de.__class__.__qualname__
        if __debug__:
            logger.critical('Missing render implementation for "%s"', cls)
            return -1
        else:
            raise NotImplementedError(f'Canvas: Missing render implementation for {cls}')

    @render.register
    def render_MultiLine(self, de: diagram.MultiLine, *, extratags: typing.Sequence[str] = tuple()) -> int:
        assert all(t not in de.tags for t in extratags)  # Do not alias "real" tags
        return self.cv.create_line(
            [p.tuple for p in de.points],
            tags=(*de.tags, *extratags),
            **tkcanvas_smooth(de.smooth),
            **tkcanvas_arrow(de.arrow),
            capstyle=de.cap.value, joinstyle=de.join.value,
            # Dash
            **tkcanvas_dash(dash=de.dash,
                            activedash=de.dashActive,
                            disableddash=de.dashDisabled),
            # Width
            **tkcanvas_width(width=de.width,
                             activewidth=de.widthActive,
                             disabledwidth=de.widthDisabled),
            # Colours
            **tkcanvas_colors(de.color, TKCANVAS_COLORMAP_MULTILINE,
                              active=de.colorActive,
                              disabled=de.colorDisabled)
        )

    @render.register
    def render_Polygon(self, de: diagram.Polygon, *, extratags: typing.Sequence[str] = tuple()) -> int:
        assert all(t not in de.tags for t in extratags)  # Do not alias "real" tags
        return self.cv.create_polygon(
            [p.tuple for p in de.points],
            tags=(*de.tags, *extratags),
            **tkcanvas_smooth(de.smooth),
            # Dash
            **tkcanvas_dash(dash=de.dash,
                            activedash=de.dashActive,
                            disableddash=de.dashDisabled),
            # Width
            **tkcanvas_width(width=de.width,
                             activewidth=de.widthActive,
                             disabledwidth=de.widthDisabled),
            # Colours
            **tkcanvas_colors(de.color,
                              active=de.colorActive,
                              disabled=de.colorDisabled)
        )

    @render.register
    def render_Rectangle(self, de: diagram.Rectangle, *, extratags: typing.Sequence[str] = tuple()) -> int:
        assert all(t not in de.tags for t in extratags)  # Do not alias "real" tags
        return self.cv.create_rectangle(
            de.topleft.tuple, de.botright.tuple,
            tags=(*de.tags, *extratags),
            # Dash
            **tkcanvas_dash(dash=de.dash,
                            activedash=de.dashActive,
                            disableddash=de.dashDisabled),
            # Width
            **tkcanvas_width(width=de.width,
                             activewidth=de.widthActive,
                             disabledwidth=de.widthDisabled),
            # Colours
            **tkcanvas_colors(de.color,
                              active=de.colorActive,
                              disabled=de.colorDisabled)
        )

    @render.register
    def render_Ellipse(self, de: diagram.Ellipse, *, extratags: typing.Sequence[str] = tuple()) -> int:
        assert all(t not in de.tags for t in extratags)  # Do not alias "real" tags
        return self.cv.create_oval(
            de.topleft.tuple, de.botright.tuple,
            tags=(*de.tags, *extratags),
            # Dash
            **tkcanvas_dash(dash=de.dash,
                            activedash=de.dashActive,
                            disableddash=de.dashDisabled),
            # Width
            **tkcanvas_width(width=de.width,
                             activewidth=de.widthActive,
                             disabledwidth=de.widthDisabled),
            # Colours
            **tkcanvas_colors(de.color,
                              active=de.colorActive,
                              disabled=de.colorDisabled)
        )

    @render.register
    def render_Text(self, de: diagram.Text, *, extratags: typing.Sequence[str] = tuple()) -> int:
        assert all(t not in de.tags for t in extratags)  # Do not alias "real" tags
        # Type Checking: `angle` is missing
        #                See https://github.com/python/typeshed/pull/10404
        return self.cv.create_text(
            list(de.point.tuple), text=de.text,
            tags=(*de.tags, *extratags),
            # # Alignment
            anchor=de.anchor.value, justify=de.justify.value,
            **tkcanvas_text_angle(de.angle),
            # Font Configuration
            font=de.font,
            # Colours
            **tkcanvas_colors(de.color, TKCANVAS_COLORMAP_TEXT,
                              active=de.colorActive,
                              disabled=de.colorDisabled)
        )



'''
Auto Layout capabilities.
'''
import logging
import typing
import math
import sys

import tkinter as tk

from . import model

logger = logging.getLogger(__name__)

# Automatic Layout
# - HORIZONTAL
# - VERTICAL
AUTO: str = 'auto'
HORIZONTAL = tk.HORIZONTAL
VERTICAL = tk.VERTICAL

LAYOUT_SYNONYMS: typing.Mapping[str, str] = {
    HORIZONTAL: '1x',  # 1 Row
    VERTICAL: 'x1',  # 1 Column
    AUTO: 'x',  # Square
}
# Multiples
# - Use the direction names directly? With a prefix?
LAYOUT_MULTIPLE: typing.Mapping[str, model.Direction] = {
    'R': model.Direction.E,
    'r': model.Direction.W,
    'H': model.Direction.H,
    'C': model.Direction.S,
    'c': model.Direction.N,
    'V': model.Direction.V,
}


def parse_amount(value):
    ''''''  # Internal, do not document
    # Allow 'x' to be replaced with the remaining widgets
    if value == 'x':
        return None
    else:
        int_value = int(value)
        if int_value == 0:
            raise ValueError('Layout Multiples: Invalid Value: %s' % value)
        return int_value


def do(layout_input: typing.Optional[str], amount: int) -> typing.Tuple[typing.Optional[str], typing.Iterable[model.GridCoordinates]]:
    # Pre-Process
    layout = None
    args: typing.Iterable[model.GridCoordinates] = iter([])
    if layout_input:
        layout = LAYOUT_SYNONYMS.get(layout_input, layout_input)
        assert layout is not None, f'Invalid Layout: {layout_input}'
    if __debug__:
        logger.debug('Final Layout: %s', layout)
    # Do it!
    if layout:
        direction_names = tuple(d.name for d in model.Direction)
        if layout.startswith(('x', 'X')):
            auto_separator = layout[0]
            if layout[1:] in ('', *direction_names):
                square = math.ceil(math.sqrt(amount))
                logger.debug('Layout: Square (%d)', square)
                layout = layout.replace(auto_separator, '%d%s%d' % (square, auto_separator, square), 1)
                logger.debug('      : %s', layout)
        if layout.startswith(tuple(LAYOUT_MULTIPLE)):
            _type = layout[0]
            multiple_direction = LAYOUT_MULTIPLE[_type]
            logger.debug('Layout: Multiples %s: %s', _type, multiple_direction)
            amounts = [parse_amount(v) for v in layout[1:].split(',')]
            args = multiple_direction.multiples(*amounts, amount=amount)
        elif 'x' in layout or 'X' in layout:
            auto_direction = model.Direction.S  # Default automatic Direction
            if any(layout.endswith(name) for name in direction_names):
                # Requires the Direction name to be 1 character long
                auto_direction = model.Direction[layout[-1]]
                layout = layout[:-1]
            auto_force: bool = 'X' in layout
            assert isinstance(auto_direction, model.Direction)
            if auto_force:
                auto_separator = 'X'
                assert 'x' not in layout
            else:
                auto_separator = 'x'
                assert 'X' not in layout
            rows, cols = [None if v == '' else int(v) for v in layout.split(auto_separator)]
            # At least one of `rows`/`cols` is not `None`
            if rows is None:
                assert cols is not None
                rows = math.ceil(amount / cols)
            if cols is None:
                assert rows is not None
                cols = math.ceil(amount / rows)
            grid_missing = rows * cols - amount
            if grid_missing > 0 and not auto_force:
                if grid_missing >= cols:
                    dcols = grid_missing // cols
                    logger.debug('      : -%d Columns', dcols)
                    cols -= dcols
                    grid_missing = rows * cols - amount
                    if __debug__:
                        if layout_input not in LAYOUT_SYNONYMS:
                            # This might be a spurious warning
                            logger.warning('%s: Non-automatic layout being unsquared: %d cols', layout_input, dcols)
                if grid_missing >= rows:
                    drows = grid_missing // rows
                    logger.debug('      : -%d Rows', drows)
                    rows -= drows
                    grid_missing = rows * cols - amount
                    if __debug__:
                        if layout_input not in LAYOUT_SYNONYMS:
                            # This might be a spurious warning
                            logger.warning('%s: Non-automatic layout being unsquared: %d rows', layout_input, drows)
            logger.debug('Layout: Automatic Grid (%d%s%d%s)[%+d]', rows, auto_separator, cols, auto_direction.name, grid_missing)
            args = auto_direction.grid(rows, cols, amount=amount,
                                       auto_fill=not auto_force)
        container_matrix = None  # For debug
        if __debug__:
            try:
                from defaultlist import defaultlist  # type: ignore
                container_matrix = defaultlist(lambda: defaultlist())
            except ImportError:
                pass  # Don't use if it doesn't exist
            logged_args = []  # Consume the iterator for logging ...
            for idx, arg in enumerate(args):
                logged_args.append(arg)
                if container_matrix is not None:  # Fill the widget locations
                    for drow in range(arg.rowspan):
                        for dcol in range(arg.columnspan):
                            container_matrix[arg.row + drow][arg.column + dcol] = idx
                else:
                    logger.debug('| %s' % arg)
            if container_matrix is not None:
                for r in container_matrix:  # Print widget locations
                    assert isinstance(r, typing.Sequence)
                    logger.debug('| %s', ' '.join(('x' * 2 if i is None else '%02d' % i for i in r)))
            args = iter(logged_args)  # ... and return a new iterator
    return layout, args


# TODO: Return a generator?
def gnested(nlayout: typing.Union[str, typing.Sequence[str]]) -> typing.Tuple[str, typing.Union[str, typing.Sequence[str]]]:
    '''Calculate the current and next layout, for nested widgets.

    Args:
        nlayout: The current layout.

    Returns:
        A tuple of ``this_layout`` (the current layout), and ``next_layout`` (the next layout level).
        ``next_layout`` is the input for this function on the next level.
    '''
    this_layout: str
    next_layout: typing.Union[str, typing.Sequence[str]]
    if isinstance(nlayout, str):
        this_layout = next_layout = nlayout
    else:
        if len(nlayout) > 0:
            this_layout = nlayout[0]
            next_layout = nlayout[1:]
        else:
            this_layout = next_layout = AUTO
    return this_layout, next_layout


if __name__ == '__main__':
    # import sys
    import argparse
    parser = argparse.ArgumentParser(description='tkmilan: Auto Layout')
    parser.add_argument('-v', '--verbose', dest='loglevel',
                        action='store_const', const=logging.DEBUG, default=logging.INFO,
                        help='Add more details to the standard error log')
    parser.add_argument('layout',
                        help='Layout String')
    parser.add_argument('amount', type=int,
                        help='Layout Amount')
    args = parser.parse_args()

    logging.basicConfig(level=args.loglevel,
                        format='%(levelname)5.5s:%(funcName)s: %(message)s',
                        stream=sys.stderr)
    layout, lst = do(args.layout, args.amount)
    logger.info('Layout: %s', layout)



import unittest
import logging

from textwrap import dedent

from tkmilan import model
from tkmilan import parser


logger = logging.getLogger(__name__)


class Test_LTML(unittest.TestCase):
    # # For copy-paste under "result"
    # print('')
    # print('golden = [')
    # for cmd in result:
    #     print(f'    {cmd},')
    # print(']')
    # print('')
    def test_simple(self):
        html_string = 'Text<a>Link</a>More Text<br/>Line Break above<a href="clean">Clean Link</a>S1<span>S2</span>S3'
        golden = [
            model.TextElementInline(text='Text'),
            model.TextElementInline(text='Link', tag='a', tags=['a::0']),
            model.TextElementInline(text='More Text'),
            model.TextElement_br(),
            model.TextElementInline(text='Line Break above'),
            model.TextElementInline(text='Clean Link', tag='a', tags=['a-clean', 'a::1']),
            model.TextElementInline(text='S1'),
            model.TextElementInline(text='S2', tag='span', tags=['span::0']),
            model.TextElementInline(text='S3'),
        ]

        result = parser.parse_LTML(html_string)
        self.assertListEqual(result, golden)

    def test_data(self):
        html_string = dedent('''<span data-one="1" data-two="2">Text</span>''')
        golden = [
            model.TextElementInline(text='Text', tag='span', tags=['span::0'], data={
                'one': '1',
                'two': '2',
            }),
        ]

        result = parser.parse_LTML(html_string)
        self.assertListEqual(result, golden)

    def test_newlines(self):
        html_string = 'Some Text\nLine Break above'
        with self.assertWarns(Warning):
            golden = [
                model.TextElementInline(text='Some Text\nLine Break above'),
            ]

            result = parser.parse_LTML(html_string)
        self.assertListEqual(result, golden)

    def test_br(self):
        html_string = dedent('''Text<br/>Text''')
        golden = [
            model.TextElementInline(text='Text'),
            model.TextElement_br(),
            model.TextElementInline(text='Text'),
        ]
        result = parser.parse_LTML(html_string)

        self.assertListEqual(result, golden)


if __name__ == '__main__':
    import sys
    logs_lvl = logging.DEBUG if '-v' in sys.argv else logging.INFO
    logging.basicConfig(level=logs_lvl, format='%(levelname)5.5s:%(funcName)s: %(message)s', stream=sys.stderr)
    unittest.main()



'''Models to store diagram information.

Implements Euclidian geometry figures in 2D, with discrete pixel coordinates
(some functions accept or provide real numbers, when integer conversion is too
lossy). Mostly implemented as `dataclasses`.

The coordinate system is common to several implementations, origin on the
topleft corner, axes directed to the rigth and down.
See :tkinter_effbot:`canvas coordinate systems <canvas.htm#coordinate-systems>`
documentation.
See also the correspoding `SVG documentation
<https://developer.mozilla.org/en-US/docs/Web/SVG/Tutorial/Positions>`_, as
example.

See also :doc:`tkmilan.model`.
'''
# Technically, coordinates can be given as strings with physical distances.
# Probably not worth it to support this.
import typing
import warnings
import logging
from functools import singledispatchmethod
import abc
from enum import Enum
from dataclasses import dataclass, field
import math
import numbers
import tkinter as tk

from . import model

logger = logging.getLogger(__name__)


# Geometric Constructs
@dataclass(frozen=True)
class Vector:
    '''A vector with integer coordinates.

    Args:
        dx: The ``x`` coordinate.
        dy: The ``y`` coordinate.

    .. automethod:: __neg__

    See Also:
        See other geometric constructs like `XY`, and `GeometryLine`.
    '''
    dx: int
    dy: int

    def __post_init__(self):
        assert self.dx == int(self.dx), f'{self}: Only integer X coordinates are supported'
        assert self.dy == int(self.dy), f'{self}: Only integer Y coordinates are supported'

    @property
    def tuple(self) -> typing.Tuple[int, int]:
        '''Represent the vector as a coordinate tuple.

        Useful to actually pass this information to most functions.
        '''
        return (self.dx, self.dy)

    def size(self) -> numbers.Real:
        '''Calculate the vector magnitude, that it, it's length.

        Returns:
            The vector magnitude, a real number.
        '''
        return typing.cast(numbers.Real, math.hypot(self.dx, self.dy))

    def scale(self, both: typing.Optional[numbers.Real] = None, *,
              sx: typing.Optional[numbers.Real] = None, sy: typing.Optional[numbers.Real] = None,
              ) -> 'Vector':
        '''Calculate a scaled vector.

        Note the scaling factors do not need to be integers, they are coerced
        to integers after the calculation using `int`.

        Choose to either scale both coordinates by the same factor using
        ``both``, or individually using ``sx``/``sy``.
        Do not combine both methods, or results might be unpredictable.
        When not given, the other coordinates remain the same.

        Args:
            both: Scale both coordinates by this factor.
            sx: Scale the ``x`` coordinate.
            sy: Scale the ``y`` coordinate.

        Returns:
            A new `Vector` object.
        '''
        assert both is None or (sx is None and sy is None), 'Scale both coordinates, or one at a time'
        return self.__class__(dx=int(self.dx * (sx or both or 1)),
                              dy=int(self.dy * (sy or both or 1)))

    def __neg__(self) -> 'Vector':
        '''Calculate the negated vector.

        Use as ``-vector``.

        Returns:
            A new `Vector` with negated coordinates.

        Note:
            This should be similar to ``vector.scale(-1)``.
        '''
        return self.__class__(dx=-self.dx, dy=-self.dy)


def VectorH(dx: int, *args, **kwargs) -> Vector:
    '''An horizontal `Vector`.

    Requires only the ``x`` coordinate.

    See Also:
        Wrapper for `Vector`.
    '''
    assert dx == int(dx)
    return Vector(dx, 0, *args, **kwargs)


def VectorV(dy: int, *args, **kwargs) -> Vector:
    '''A vertical `Vector`.

    Requires only the ``y`` coordinate.

    See Also:
        Wrapper for `Vector`.
    '''
    assert dy == int(dy)
    return Vector(0, dy, *args, **kwargs)


def VectorPolar(r: numbers.Real, theta: numbers.Real) -> Vector:
    '''A "polar" `Vector`, defined by radius and angle.

    See `Wikipedia polar coordinates
    <https://en.wikipedia.org/wiki/Polar_coordinate_system>`_ information.

    Args:
        r: Radius
        theta: Vector angle, in degrees.
    '''
    # Remember: Y axis is the other way around
    return Vector(
        int(r * math.cos(math.radians(theta))),
        -int(r * math.sin(math.radians(theta))),
    )


@dataclass(frozen=True)
class XY:
    '''A location with integer coordinates.

    Args:
        x: The ``x`` coordinate.
        y: The ``y`` coordinate.

    .. automethod:: __add__
    .. automethod:: __sub__

    See Also:
        See other geometric constructs like `Vector`, and `GeometryLine`.
    '''
    x: int
    y: int

    @property
    def tuple(self) -> typing.Tuple[int, int]:
        '''Represent the vector as a coordinate tuple.

        Useful to actually pass this information to most functions.
        '''
        return (self.x, self.y)

    def vto(self, other: 'XY') -> Vector:
        '''Calculate the `Vector` joining this and ``other`` point.

        The result is a vector that when added to this point, results in the
        ``other`` point.

        Args:
            other: The other `XY`, see the description above.

        Returns:
            A new `Vector` object.
        '''
        return Vector(other.x - self.x, other.y - self.y)

    def edistance(self, other: 'XY') -> numbers.Real:
        '''Calculate the Euclidian distance between this and ``other`` point.

        Returns:
            The distance between both points, as a floating point number.
        '''
        return typing.cast(numbers.Real, math.hypot(
            abs(other.x - self.x),
            abs(other.y - self.y),
        ))

    @singledispatchmethod
    def __add__(self, obj: typing.Any):
        '''Calculate the sum of the given object with this location.

        Supported Object Types:

        - `Vector`

        Use as ``xy + vector``.
        '''
        raise NotImplementedError

    @__add__.register
    def __add__Vector(self, obj: Vector):
        return self.__class__(x=self.x + obj.dx, y=self.y + obj.dy)

    @singledispatchmethod
    def __sub__(self, obj: typing.Any):
        '''Calculate the subtraction of the given object with this location.

        Supported Object Types:

        - `Vector`

        Use as ``xy - vector``.
        '''
        raise NotImplementedError

    @__sub__.register
    def __sub__Vector(self, obj: Vector):
        return self + (-obj)


@dataclass(frozen=True)
class GeometryLine:
    '''A geometric line (not an element like `Line`), for easing some
    calculations. Uses real parameters.

    To create this object based on other inputs, see `points` and
    `point_slope`.

    Args:
        m: Line slope. `None` means a vertical line (equivalent to infinite
            slope).
        c0: For vertical lines, it's the ``x`` coordinate for all points.
            For other lines, it's the ``y`` coordinate at ``x=0``.
            See `equation` for the complete picture.

    See Also:
        See other geometric constructs like `XY`, and `Vector`.

    .. automethod:: __contains__
    '''
    m: typing.Optional[numbers.Real]
    c0: numbers.Real

    def __init__(self,
                 m: typing.Union[int, float, numbers.Real, None],
                 c0: typing.Union[int, float, numbers.Real],
                 ):
        object.__setattr__(self, 'm', typing.cast(typing.Optional[numbers.Real], None if m is None else float(m)))
        object.__setattr__(self, 'c0', typing.cast(numbers.Real, float(c0)))

    def __str__(self):
        return f'{self.__class__.__qualname__}<{self.equation}>'

    @property
    def equation(self) -> str:
        '''Derive an equation for the line.

        This uses the Slope-intercept form.
        '''
        if self.m is None:
            # Vertical
            return f'x = {self.c0}'
        elif self.m == 0.0:
            return f'y = {self.c0}'
        else:
            string = 'y = '
            if self.m == 1.0:
                string += 'x'
            elif self.m == -1.0:
                string += '-x'
            else:
                string += f'{self.m} . x'
            if self.c0 < typing.cast(numbers.Real, 0.0):
                string += '+ {self.c0}'
            elif self.c0 > typing.cast(numbers.Real, 0.0):
                string += '- {-self.c0}'
            else:
                assert self.c0 == 0.0
            return string

    @singledispatchmethod
    def __contains__(self, obj: typing.Any) -> bool:
        '''Calculate if the given object is contained on this line.

        Supported Object Types:

        - `XY`

        Use as ``xy in line``.
        '''
        raise NotImplementedError

    @__contains__.register
    def __contains__XY(self, obj: XY):
        if self.m is None:
            # Vertical
            return obj.x == self.c0
        else:
            return math.isclose(obj.y, self.m * obj.x + self.c0)

    def intersect_y(self, y: int) -> typing.Optional[XY]:
        '''Calculate the intersection of this line with an horizontal line.

        Args:
            y: The ``y`` coordinate for the horizontal line.

        Returns:
            Return the intersection point, if exists.
            Returns `None` when there is not intersection, or there is an
            infinite number of points.

        See Also:
            See `intersect_line` for calculating the intersection with any
            line.
        '''
        if self.m is None:
            return XY(int(self.c0), y)
        elif self.m == 0.0:
            return None
        else:
            return XY(int((y - self.c0) / self.m), y)

    def intersect_x(self, x: int) -> typing.Optional[XY]:
        '''Calculate the intersection of this line with a vertical line.

        Args:
            x: The ``x`` coordinate for the horizontal line.

        Returns:
            Return the intersection point, if exists.
            Returns `None` when there is not intersection, or there is an
            infinite number of points.

        See Also:
            See `intersect_line` for calculating the intersection with any
            line.
        '''
        if self.m is None:
            return None
        else:
            return XY(x, int(self.m * x + self.c0))

    def intersect_line(self, other: 'GeometryLine') -> 'typing.Optional[XY]':
        '''Calculate the intersection of this line with any other line.

        Args:
            other: The other line.

        Returns:
            Return the intersection point, if exists.
            Returns `None` when there is not intersection, or there is an
            infinite number of points.

        See Also:
            See `intersect_x` and `intersect_y` for calculating the
            intersection with special lines.
        '''
        if self.m == other.m:
            # Parallel
            if __debug__:
                if self.c0 == other.c0:
                    logger.warning('%s|%s: Both lines are the same', self, other)
            return None
        elif other.m is None:
            # other + Vertical
            return self.intersect_x(int(other.c0))
        elif self.m is None:
            # Vertical + other
            return other.intersect_x(int(self.c0))
        else:
            assert self.m is not None and other.m is not None
            m_delta = self.m - other.m
            return XY(
                x=int((other.c0 - self.c0) / m_delta),
                y=int((other.c0 * self.m - other.m * self.c0) / m_delta),
            )

    def parallel_point(self, p: XY) -> 'GeometryLine':
        '''Calculate the parallel line that goes through a given point.

        There is only a single line parallel to the current one that goes
        through an external point.
        If the point is already in the current line, returns itself.

        Args:
            p: A point on the new line.
        '''
        if p in self:
            if __debug__:
                logger.warning('%s: Point %s in line already', self, p)
            return self
        elif self.m is None:
            # Vertical
            return self.__class__(m=None, c0=p.x)
        else:
            hdelta = p.y - (self.m * p.x + self.c0)
            return self.__class__(m=self.m, c0=self.c0 + hdelta)

    def perpendicular_point(self, p: XY) -> 'GeometryLine':
        '''Calculate the perpendicular line that goes through a given point.

        There is only a single line perpendicular to the current one that goes
        through any point.

        Args:
            p: A point on the new line.
        '''
        if self.m is None:
            # Vertical ⟂ Horizontal
            return self.__class__(0.0, p.y)
        elif self.m == 0.0:
            # Horizontal ⟂ Vertical
            return self.__class__(None, p.x)
        else:
            pm = math.pow(-self.m, -1.0)  # (- 1 / m)
            return self.__class__.point_slope(p, typing.cast(numbers.Real, pm))

    # TODO: On Python 3.11:: -> typing.Self
    @classmethod
    def points(cls, p1: XY, p2: XY) -> 'GeometryLine':
        '''Create the line that goes through both points.

        If both points refer to the same location, this is an error.

        Args:
            p1: One of the points in the line.
            p2: One of the points in the line.
        '''
        assert p1 != p2, f'Same point, no line exists: {p1}/{p2}'
        if p1.x == p2.x:
            # Vertical
            return cls(m=None, c0=float(p1.x))
        elif p1.y == p2.y:
            # Horizontal
            return cls(m=0.0, c0=float(p2.y))
        else:
            return cls(
                m=(p1.y - p2.y) / (p1.x - p2.x),
                c0=(p1.x * p2.y - p2.x * p1.y) / (p1.x - p2.x)
            )

    # TODO: On Python 3.11:: -> typing.Self
    @classmethod
    def point_slope(cls, p: XY, m: typing.Optional[numbers.Real]) -> 'GeometryLine':
        '''Create the line that goes through a point with the given slope.

        Args:
            p: One of the points in the line.
            m: Line slope. `None` means infinite slope, that is, a vertical
                line.
        '''
        if m is None:
            # Vertical
            return cls(m=None, c0=float(p.x))
        elif m == 0:
            # Horizontal
            return cls(m=0.0, c0=float(p.y))
        else:
            return cls(m=float(m), c0=float(p.y - m * p.x))


# Settings Objects
@dataclass
class C:
    '''Colour settings.

    The colours are :tk_lib:`arbitrary colours <GetColor.htm#M7>` with the
    format ``#RRGGBB``/``#RGB``, or the following :tk:`recognized colour names
    <colors.htm>`.

    Args:
        fill: The filling colour, for the internal element area.
            Optional.
        outline: The outline colour, for the line that marks the element
            boundary. Optional.
    '''
    fill: typing.Optional[str] = None
    outline: typing.Optional[str] = None

    @property
    def reverse(self) -> 'C':
        '''Calculate reversed colour settings.

        To clarify, this is a setting when the filling colour is the outline
        colour, and vice versa.
        '''
        return C(fill=self.outline, outline=self.fill)

    def w(self, **kwargs: typing.Optional[str]) -> 'C':
        '''Calculate a colour setting with different values.

        Doesn't make sense to create a new setting with both new colours, this
        is equivalent to create a new object. This is only useful to create a
        new setting changing only a single colour.

        Args:
            fill: The new filling colour. Optional.
            outline: The new outline colour. Optional.
        '''
        assert len(kwargs) == 1, 'Change a single colour at the time'
        # Equivalent to using sentinel values as defaults, `None` is a valid value
        return C(fill=kwargs.get('fill', self.fill), outline=kwargs.get('outline', self.outline))


@dataclass
class A:
    '''Arrow settings.

    Check a visual representation of the arrow head shape on the ``arrowshape``
    section in ``Tk`` :tkinter_nmt:`NMU canvas create line <create_line.html>`.

    Make sure to select one of ``atStart``/``atEnd``, otherwise this represents
    no arrow at all.

    Args:
        d1: Arrow Head Shape: Distance along the line, from neck to tip.
        d2: Arrow Head Shape: Distance along the line, from trailing points to
            tip.
        d3: Arrow Head Shape: Distance normal from the line, from the outside
            edge to the trailing points.
        atStart: Include an arrow head on the start of the line.
            Defaults to `False`.
        atEnd: Include an arrow head on the end of the line.
            Defaults to `True`.
    '''
    d1: int = 8
    d2: int = 10
    d3: int = 3
    atStart: bool = False
    atEnd: bool = True


@dataclass
class D:
    '''Dash settings.

    The ``pattern`` is a tuple of distances to consider. Only odd-indexed
    pattern distances are drawn, the even-indexed distances are skipped (or
    "drawn" with transparent color). There's also an initial ``offset``, can
    even be negative.

    Check as visual representation of the dash patterns in ``Tk``
    :tkinter_nmt:`NMT canvas dash patterns <dash-patterns.html>` or :tk:`canvas
    dash patterns <canvas.htm#M26>` documentation.

    Args:
        pattern: Tuple with distances to be considered.
            See the description above on how this is represented.
        offset: Offset on pattern start.
            Defaults to ``0``, no offset.
    '''
    pattern: typing.Tuple[int, ...]
    offset: int = 0

    def __init__(self, *pattern: int, offset: int = 0):
        self.pattern = tuple(pattern)
        assert len(self.pattern) > 0
        self.offset = offset


class Cap(Enum):
    '''Cap settings.

    Check a visual representation of the style in ``Tk`` :tkinter_nmt:`NMT cap
    <cap-join-styles.html>` documentation.
    '''
    BUTT = tk.BUTT
    '''The end of the line is cut off square at a line that passes through the
    endpoint.'''
    PROJECTING = tk.PROJECTING
    '''The end of the line is cut off square, but the cut line projects past
    the endpoint a distance equal to half the line's width.'''
    ROUND = tk.ROUND
    '''The end describes a semicircle centered on the endpoint.'''


class Join(Enum):
    '''Join settings.

    Check a visual representation of the style in ``Tk`` :tkinter_nmt:`NMT join
    <cap-join-styles.html>` documentation.
    '''
    ROUND = tk.ROUND
    '''The join is a circle centered on the point where the adjacent line segments meet.'''
    BEVEL = tk.BEVEL
    '''A flat facet is drawn at an angle intermediate between the angles of the adjacent lines.'''
    MITER = tk.MITER
    '''The edges of the adjacent line segments are continued to meet at a sharp point.'''


class SmoothAlgorithm(Enum):
    '''Line smoothing algorithm selection.

    There is no Python documentation, see ``Tk`` :tk:`canvas line smooth
    <canvas.htm#M143>` documentation.

    See Also:
        See `Smooth` for the usable line smoothing settings.
    '''
    BEZIER2 = 'bezier'
    '''Draw line as series of quadratic Bézier curves.

    See `Wikipedia quadratic Bézier curves
    <https://en.wikipedia.org/wiki/B%C3%A9zier_curve#Quadratic_curves>`_
    information.
    '''
    BEZIER3 = 'raw'
    '''Draw line as series of cubic Bézier curves.

    See `Wikipedia cubic Bézier curves
    <https://en.wikipedia.org/wiki/B%C3%A9zier_curve#Higher-order_curves>`_
    information.
    '''


@dataclass
class Smooth:
    '''Line smoothing settings.

    There is no Python documentation, see ``Tk`` :tk:`canvas line smooth
    <canvas.htm#M143>` documentation.

    Args:
        algorithm: Line smoothing algorithm.
            Defaults to cubic Bézier curves.
        steps: The curve is rendered as a series of line segments.
            This is the amount of lines to render. Defaults to 12.
    '''
    algorithm: SmoothAlgorithm = SmoothAlgorithm.BEZIER3
    steps: int = 12

    def __post_init__(self):
        assert self.steps > 0, '{self}: Number of steps must be positive'


# Diagram Elements
class DiagramElement(abc.ABC):
    '''Common diagram element class.

    This is only an abstract class, a common base class for all the possible
    diagram elements.
    '''
    def iterate(self) -> 'typing.Iterator[DiagramElementSingle]':
        '''Gather all constituent `DiagramElementSingle` objects, recursively.'''
        if isinstance(self, DiagramElementSingle):
            yield self
        elif isinstance(self, DiagramElementMultiple):
            for item in self.items:
                yield from self.iterate()
        else:
            raise NotImplementedError(f'Unknown Type: {self!r}')


@dataclass
class DiagramElementSingle(DiagramElement):
    '''Common diagram single element class.

    All single element classes inherit from this.
    '''
    pass


@dataclass
class DiagramElementMultiple(DiagramElement):
    '''Aggregator for multiple `DiagramElement` objects.

    Certain diagram elements only make sense together, group them here.
    '''
    items: typing.Sequence[DiagramElement]

    def __init__(self, *items: DiagramElement):
        self.items = tuple(items)


# Support arbitrary layers?
class Diagram(abc.ABC):
    '''Diagram generator class.

    This should be subclassesed and it's function implemented with `iter` that
    create the diagram itself, based on possible ``__init__`` state, and also
    per-function arguments.

    There are three available "layers" for elements, each built from a
    function, in order:

    - `setup_bg_b`: Background Back (``z=-1``).
    - `setup_fg`: Foreground (``z=0``)
    - `setup_bg_f`: Background Front (``z=1``).

    All functions are optional, they default to generate no elements.

    This is only an abstract class, a common base class for all the possible
    diagrams.
    '''
    # TODO: Implement arbitrary layers?
    # No `__init__`, this is specific for each one
    BACKGROUND: typing.Optional[str] = None
    '''Canvas background colour.

    Optional. See `C` for colour specification.
    '''
    DISABLEDBACKGROUND: typing.Optional[str] = None
    '''Canvas background colour, when disabled.

    Optional. See `C` for colour specification.
    '''
    MIN_SIZE: typing.Tuple[typing.Optional[int], typing.Optional[int]] = (None, None)
    '''Canvas minimum size for rendering.

    If the canvas size is smaller than any of the (non-`None`) sizes, rendering
    is skipped.

    Optional.
    '''

    def setup_bg_b(self, *, cwidth: int, cheight: int) -> typing.Iterator[DiagramElement]:
        '''Draw the Background Back layer.

        Should be a generator creating a finite amount of `DiagramElement`.
        '''
        # Drawn only when size changes
        return iter([])

    def setup_fg(self, *, cwidth: int, cheight: int) -> typing.Iterator[DiagramElement]:
        '''Draw the Foreground layer.

        Should be a generator creating a finite amount of `DiagramElement`.
        '''
        # Drawn always
        return iter([])

    def setup_bg_f(self, *, cwidth: int, cheight: int) -> typing.Iterator[DiagramElement]:
        '''Draw the Background Front layer.

        Should be a generator creating a finite amount of `DiagramElement`.
        '''
        # Drawn only when size changes
        return iter([])


# # External Diagram Elements

# # # Base Elements
@dataclass
class MultiLine(DiagramElementSingle):
    '''Diagram Element: multi-segment line.

    Represent multiple line segments with common points: line segment between
    first and second, second and third, etc...

    There is no Python documentation, see ``Tk`` :tk:`canvas common options
    <canvas.htm#M99>` and :tk:`canvas line options <canvas.htm#M143>`. See also
    :tkinter_nmt:`NMT create line <create_line.html>` documentation.

    Args:
        points: Sequence of points to draw the line segments.
            Must have at least two points.
        color: Colour for line segments, see `C` for colour specification.
            Considers only ``outline``.
        width: Width for line segments, in pixels.
            Optional, defaults to minimum line width.
        dash: Dashed line segments setting.
            Optional, defaults to solid line.
        arrow: Arrow line settings.
            Optional, defaults to no arrows anywhere.
        cap: Line capping setting. Defaults to `Cap.BUTT`.
        join: Line join setting. Defaults to `Join.ROUND`.
        smooth: Line smoothing setting.
            Optional, defaults to straight line.
        colorActive: ``color`` for active elements, that is, on mouse hover.
        widthActive: ``width`` for active elements, that is, on mouse hover.
        dashActive: ``dash`` for disabled elements.
            ``offset`` is not supported.
        colorDisabled: ``color`` for disabled elements.
        widthDisabled: ``width`` for disabled elements.
        dashDisabled: ``dash`` for disabled elements.
            ``offset`` is not supported.
        tags: Sequence of tags to apply to the element.
            Must not include any "internal tags", starting with ``:``.
    '''
    # - No stipple support, it's not cross platform
    points: typing.Sequence[XY]
    color: C = field(default_factory=C)  # TODO: Accept single string, `C(outline=x)`
    width: typing.Optional[int] = None
    dash: typing.Optional[D] = None  # TODO: Accept (int, int), `D(*x)`
    arrow: typing.Optional[A] = None
    cap: Cap = Cap.BUTT
    join: Join = Join.ROUND
    smooth: typing.Optional[Smooth] = None
    colorActive: C = field(default_factory=C)  # TODO: Accept single string, `C(outline=x)`
    widthActive: typing.Optional[int] = None
    dashActive: typing.Optional[D] = None  # TODO: Accept (int, int), `D(*x)`
    colorDisabled: C = field(default_factory=C)  # TODO: Accept single string, `C(outline=x)`
    widthDisabled: typing.Optional[int] = None
    dashDisabled: typing.Optional[D] = None  # TODO: Accept (int, int), `D(*x)`
    tags: typing.Sequence[str] = tuple()

    if __debug__:
        # Validate settings
        def __post_init__(self):
            if len(self.points) < 2:
                raise ValueError(f'{self}: A Line has at least two points')
            if self.smooth is None:
                if len(set(self.points)) != len(self.points):
                    logger.warning('%s: Repeated "points"', self)
            else:
                logger.warning('%s: Smoothing algorithm selection is not fully implemented', self)
            for c in ('color', 'colorActive', 'colorDisabled'):
                cv = getattr(self, c)
                if cv.fill is not None:
                    if cv.outline is None:
                        logger.warning('%s: To paint a line with `%s`, use "outline" exclusively', self, c)
                    else:
                        logger.warning('%s: Do not use `fill=%s`, use a Polygon instead', self, cv.fill)
            for tag in self.tags:
                if tag.startswith(':'):  # Internal tags, forbid
                    raise ValueError(f'{self}: Invalid tag `{tag}`')
            for w in ('width', 'widthActive', 'widthDisabled'):
                wv = getattr(self, w)
                if wv is not None and wv <= 0:
                    raise ValueError(f'{self}: "{w}" is a distance')
            for tn in ('dash', 'dashActive', 'dashDisabled'):
                tv = getattr(self, tn)
                if tv is not None:
                    if any(v <= 0 or v > 255 for v in tv.pattern):
                        raise ValueError(f'{self}: "{tn}" distances have a range of ]0, 255]')
                    if tn != 'dash' and tv.offset != 0:
                        raise ValueError(f'{self}: "{tn}" does not support offsets, uses `dash` value')
            if self.arrow:
                if (self.arrow.atStart, self.arrow.atEnd) != (False, False):
                    raise ValueError(f'{self}: When requesting arrows, include at least one')


@dataclass
class Polygon(DiagramElementSingle):
    '''Diagram Element: polygon.

    Represent a single regular polygon.

    There is no Python documentation, see ``Tk`` :tk:`canvas common options
    <canvas.htm#M99>` and :tk:`canvas polygon options <canvas.htm#M151>`. See
    also :tkinter_nmt:`NMT create polygon <create_polygon.html>` documentation.

    Args:
        points: Sequence of points to draw the polygon.
            Must have at least three points.
        color: Colour for outline and fill area, see `C` for colour
            specification.
        width: Width for outlines, in pixels.
            Optional, defaults to minimum line width.
        dash: Dashed outline setting.
            Optional, defaults to solid line.
        join: Outline join setting. Defaults to `Join.ROUND`.
        smooth: Outline smoothing setting.
            Optional, defaults to straight line.
        colorActive: ``color`` for active elements, that is, on mouse hover.
        widthActive: ``width`` for active elements, that is, on mouse hover.
        dashActive: ``dash`` for disabled elements.
            ``offset`` is not supported.
        colorDisabled: ``color`` for disabled elements.
        widthDisabled: ``width`` for disabled elements.
        dashDisabled: ``dash`` for disabled elements.
            ``offset`` is not supported.
        tags: Sequence of tags to apply to the element.
            Must not include any "internal tags", starting with ``:``.
    '''

    points: typing.Sequence[XY]
    color: C = field(default_factory=C)
    width: typing.Optional[int] = None
    dash: typing.Optional[D] = None  # TODO: Accept (int, int), `D(*x)`
    join: Join = Join.ROUND
    smooth: typing.Optional[Smooth] = None
    colorActive: C = field(default_factory=C)
    widthActive: typing.Optional[int] = None
    dashActive: typing.Optional[D] = None  # TODO: Accept (int, int), `D(*x)`
    colorDisabled: C = field(default_factory=C)
    widthDisabled: typing.Optional[int] = None
    dashDisabled: typing.Optional[D] = None  # TODO: Accept (int, int), `D(*x)`
    tags: typing.Sequence[str] = tuple()

    if __debug__:
        # Validate settings
        def __post_init__(self):
            if len(self.points) < 3:
                raise ValueError(f'{self}: A Polygon has at least three points')
            if self.smooth is None:
                if len(set(self.points)) != len(self.points):
                    logger.warning('%s: Repeated "points"', self)
            else:
                logger.warning('%s: Smoothing algorithm selection is not fully implemented', self)
            for tag in self.tags:
                if tag.startswith(':'):  # Internal tags, forbid
                    raise ValueError(f'{self}: Invalid tag `{tag}`')
            for w in ('width', 'widthActive', 'widthDisabled'):
                wv = getattr(self, w)
                if wv is not None and wv <= 0:
                    raise ValueError(f'{self}: "{w}" is a distance')
            for tn in ('dash', 'dashActive', 'dashDisabled'):
                tv = getattr(self, tn)
                if tv is not None:
                    if any(v <= 0 or v > 255 for v in tv.pattern):
                        raise ValueError(f'{self}: "{tn}" distances have a range of ]0, 255]')
                    if tn != 'dash' and tv.offset != 0:
                        raise ValueError(f'{self}: "{tn}" does not support offsets, uses `dash` value')


@dataclass
class Rectangle(DiagramElementSingle):
    '''Diagram Element: simple rectangle, parallel to the axes.

    Represent a simple rectangle, with sides parallel to the axes. If you need
    a rotated rectangle, use a `Polygon`.

    There is no Python documentation, see ``Tk`` :tk:`canvas common options
    <canvas.htm#M99>` and :tk:`canvas rectangle options <canvas.htm#M155>`. See
    also :tkinter_nmt:`NMT create rectangle <create_rectangle.html>` documentation.

    Args:
        topleft: The top left vertex (lower X and Y coordinates).
        botright: The bottom right vertex (higher X and Y coordinates).
        color: Colour for outline and fill area, see `C` for colour
            specification.
        width: Width for outlines, in pixels.
            Optional, defaults to minimum line width.
        dash: Dashed outline setting.
            Optional, defaults to solid line.
        colorActive: ``color`` for active elements, that is, on mouse hover.
        widthActive: ``width`` for active elements, that is, on mouse hover.
        dashActive: ``dash`` for disabled elements.
            ``offset`` is not supported.
        colorDisabled: ``color`` for disabled elements.
        widthDisabled: ``width`` for disabled elements.
        dashDisabled: ``dash`` for disabled elements.
            ``offset`` is not supported.
        tags: Sequence of tags to apply to the element.
            Must not include any "internal tags", starting with ``:``.
    '''
    topleft: XY
    botright: XY
    color: C = field(default_factory=C)
    width: typing.Optional[int] = None
    dash: typing.Optional[D] = None  # TODO: Accept (int, int), `D(*x)`
    colorActive: C = field(default_factory=C)
    widthActive: typing.Optional[int] = None
    dashActive: typing.Optional[D] = None  # TODO: Accept (int, int), `D(*x)`
    colorDisabled: C = field(default_factory=C)
    widthDisabled: typing.Optional[int] = None
    dashDisabled: typing.Optional[D] = None  # TODO: Accept (int, int), `D(*x)`
    tags: typing.Sequence[str] = tuple()

    if __debug__:
        # Validate settings
        def __post_init__(self):
            deltaV = self.topleft.vto(self.botright)
            if 0 in (deltaV.dx, deltaV.dy):
                raise ValueError(f'{self}: Must have size on both axes')
            if self.topleft.x > self.botright.x or self.topleft.y > self.botright.y:
                xy_tl = XY(min(self.topleft.x, self.botright.x), min(self.topleft.y, self.botright.y))
                xy_br = XY(max(self.topleft.x, self.botright.x), max(self.topleft.y, self.botright.y))
                raise ValueError(f'{self}: review coordinates: `topleft={xy_tl}, botright={xy_br}`')
            for tag in self.tags:
                if tag.startswith(':'):  # Internal tags, forbid
                    raise ValueError(f'{self}: Invalid tag `{tag}`')
            for w in ('width', 'widthActive', 'widthDisabled'):
                wv = getattr(self, w)
                if wv is not None and wv <= 0:
                    raise ValueError(f'{self}: "{w}" is a distance')
            for tn in ('dash', 'dashActive', 'dashDisabled'):
                tv = getattr(self, tn)
                if tv is not None:
                    if any(v <= 0 or v > 255 for v in tv.pattern):
                        raise ValueError(f'{self}: "{tn}" distances have a range of ]0, 255]')
                    if tn != 'dash' and tv.offset != 0:
                        raise ValueError(f'{self}: "{tn}" does not support offsets, uses `dash` value')

    @property
    def center(self) -> XY:
        '''Calculate the center of the rectangle.

        Note that since integer coordinates are used, this might be off from
        the "real" center by at most half pixel.
        '''
        return XY((self.topleft.x + self.botright.x) // 2, (self.topleft.y + self.botright.y) // 2)


@dataclass
class Ellipse(DiagramElementSingle):
    '''Diagram Element: simple ellipse, parallel to the axes.

    Represent a simple ellipse, with sides parallel to the axes. Remember that
    a circle is a special case for an ellipse, see `CircleCenter`.

    There is no Python documentation, see ``Tk`` :tk:`canvas common options
    <canvas.htm#M99>` and :tk:`canvas oval options <canvas.htm#M150>`. See
    also :tkinter_nmt:`NMT create oval <create_oval.html>` documentation.

    Args:
        topleft: The top left vertex (lower X and Y coordinates).
        botright: The bottom right vertex (higher X and Y coordinates).
        color: Colour for outline and fill area, see `C` for colour
            specification.
        width: Width for outlines, in pixels.
            Optional, defaults to minimum line width.
        dash: Dashed outline setting.
            Optional, defaults to solid line.
        colorActive: ``color`` for active elements, that is, on mouse hover.
        widthActive: ``width`` for active elements, that is, on mouse hover.
        dashActive: ``dash`` for disabled elements.
            ``offset`` is not supported.
        colorDisabled: ``color`` for disabled elements.
        widthDisabled: ``width`` for disabled elements.
        dashDisabled: ``dash`` for disabled elements.
            ``offset`` is not supported.
        tags: Sequence of tags to apply to the element.
            Must not include any "internal tags", starting with ``:``.
    '''
    topleft: XY
    botright: XY
    color: C = field(default_factory=C)
    width: typing.Optional[int] = None
    dash: typing.Optional[D] = None  # TODO: Accept (int, int), `D(*x)`
    colorActive: C = field(default_factory=C)
    widthActive: typing.Optional[int] = None
    dashActive: typing.Optional[D] = None  # TODO: Accept (int, int), `D(*x)`
    colorDisabled: C = field(default_factory=C)
    widthDisabled: typing.Optional[int] = None
    dashDisabled: typing.Optional[D] = None  # TODO: Accept (int, int), `D(*x)`
    tags: typing.Sequence[str] = tuple()

    if __debug__:
        # Validate settings
        def __post_init__(self):
            deltaV = self.topleft.vto(self.botright)
            if 0 in (deltaV.dx, deltaV.dy):
                raise ValueError(f'{self}: Must have size on both axes')
            if self.topleft.x > self.botright.x or self.topleft.y > self.botright.y:
                xy_tl = XY(min(self.topleft.x, self.botright.x), min(self.topleft.y, self.botright.y))
                xy_br = XY(max(self.topleft.x, self.botright.x), max(self.topleft.y, self.botright.y))
                raise ValueError(f'{self}: review coordinates: `topleft={xy_tl}, botright={xy_br}`')
            for tag in self.tags:
                if tag.startswith(':'):  # Internal tags, forbid
                    raise ValueError(f'{self}: Invalid tag `{tag}`')
            for w in ('width', 'widthActive', 'widthDisabled'):
                wv = getattr(self, w)
                if wv is not None and wv <= 0:
                    raise ValueError(f'{self}: "{w}" is a distance')
            for tn in ('dash', 'dashActive', 'dashDisabled'):
                tv = getattr(self, tn)
                if tv is not None:
                    if any(v <= 0 or v > 255 for v in tv.pattern):
                        raise ValueError(f'{self}: "{tn}" distances have a range of ]0, 255]')
                    if tn != 'dash' and tv.offset != 0:
                        raise ValueError(f'{self}: "{tn}" does not support offsets, uses `dash` value')

    @property
    def center(self) -> XY:
        '''Calculate the center of the ellipse.

        Note that since integer coordinates are used, this might be off from
        the "real" center by at most half pixel.
        '''
        return XY((self.topleft.x + self.botright.x) // 2, (self.topleft.y + self.botright.y) // 2)

    @property
    def leccentricity(self) -> numbers.Real:
        '''Calculate the linear eccentricity of the ellipse.

        This is a measure of how closer is the ellipse to a circle. ``0`` means
        it's a circle, any other number less than ``1`` means it's an ellipse.

        Returns:
            The linear eccentricity of the ellipse, as a floating point number.
        '''
        a = (self.botright.x - self.topleft.x) // 2
        b = (self.botright.y - self.topleft.y) // 2
        assert a > 0 or b > 0
        if a == b:
            # Circle
            return typing.cast(numbers.Real, 0)
        else:
            # Make sure `a` is the semi-major axis
            if b > a:
                a, b = b, a
            assert a > b
            return typing.cast(numbers.Real, math.sqrt((a + b) * (a - b)))

    def foci(self) -> typing.Tuple[XY, XY]:
        '''Calculate both focus points of the ellipse

        These are the points where any outline point is equidistant from both
        points.

        Note that since integer coordinates are used, this might be off from
        the "real" foci by at most half pixel.

        See Also:
            The `center` is the midpoint between these two. They all coincide
            if the ellipse is a circle.
        '''
        # TODO: Make sure this is correct, if the ellipse is "vertical"
        vC = VectorH(int(self.leccentricity))
        return self.center - vC, self.center + vC


@dataclass
class Text(DiagramElementSingle):
    '''Diagram Element: text.

    Represent a text string anchored on a specific coordinate, possibly
    rotated.

    There is no Python documentation, see ``Tk`` :tk:`canvas common options
    <canvas.htm#M99>` and :tk:`canvas text options <canvas.htm#M156>`. See
    also :tkinter_nmt:`NMT create text <create_text.html>` documentation.

    Args:
        point: Anchor point to locate the text
        text: The text to draw
        color: Text colour, see `C` for colour specification.
            Considers only ``fill``.
        anchor: Where to anchor the text, related to ``point``.
        angle: Rotate the text, in degrees. Defaults to no rotation.
        justify: How to justify multi-line text. See `model.Justification`.
        colorActive: ``color`` for active elements, that is, on mouse hover.
        colorDisabled: ``color`` for disabled elements.
    '''
    # TODO: Create proper `Font` settings
    #       Do not document `font`, it's still a hack
    # TODO: Document how angle and achor interact.
    point: XY
    text: str
    color: C = field(default_factory=C)  # TODO: Accept single string, `C(fill=x)`
    anchor: model.CP = model.CP.center
    angle: numbers.Real = typing.cast(numbers.Real, 0.0)
    justify: model.Justification = model.Justification.Left
    font: typing.Any = None
    colorActive: C = field(default_factory=C)  # TODO: Accept single string, `C(fill=x)`
    colorDisabled: C = field(default_factory=C)  # TODO: Accept single string, `C(fill=x)`
    # TODO: Implement "lineWidth"?
    # Skip "underline"
    tags: typing.Sequence[str] = tuple()

    if __debug__:
        def __post_init__(self):
            if self.text == '':
                raise ValueError(f'{self}: Must have some text')
            for c in ('color', 'colorActive', 'colorDisabled'):
                cv = getattr(self, c)
                if cv.outline is not None and cv.fill is None:
                    logger.warning('%s: To paint text with `%s`, use "fill" exclusively', self, c)
            for tag in self.tags:
                if tag.startswith(':'):  # Internal tags, forbid
                    raise ValueError(f'{self}: Invalid tag `{tag}`')
            if self.angle < 0.0 or self.angle > 360.0:
                raise ValueError(f'{self}: Invalid angle `{self.angle}`')

# TODO: Implement `Image`
# TODO: Implement `EllipseArc` (enhance Ellipse instead?)


# # # Complex Elements
def DoubleMultiLine(*args,
                    widthSmall: int, deltaWidth: int = 2,
                    colorBig: C, colorSmall: C,
                    **kwargs,
                    ) -> DiagramElementMultiple:
    '''Draw two `MultiLine` on the same location, with different sizes.

    The "big" one is drawn first (on the back), with a larger ``deltaWidth``
    size.

    The size can be controlled using ``widthSmall`` and ``deltaWidth``.

    It's equivalent to three lines, like this::

       -------------
       BIG  size=w/D
       -------------
       small  size=w
       -------------
       BIG  size=w/D
       -------------

    This is their relation::

        w = widthSmall
        widthBig = deltaWidth * w
        D = 2 / (deltaWidth - 1)

    Examples::

        d = 2 ==> D = 2
        d = 3 ==> D = 1
        d = 4 ==> D = 2/3

    '''
    if 'color' in kwargs:
        warnings.warn('Do not use "color", use `colorBig` and `colorSmall`', stacklevel=2)
        kwargs.pop('color')
    if 'width' in kwargs:
        warnings.warn('Do not use "width", use `widthSmall` and `deltaWidth`', stacklevel=2)
        kwargs.pop('width')
    assert 'color' not in kwargs and 'width' not in kwargs
    widthBig = int(deltaWidth * widthSmall)
    if __debug__:
        logger.debug('Width: widthSmall=%d widthBig=%d', widthSmall, widthBig)
    return DiagramElementMultiple(
        MultiLine(*args, color=colorBig, width=widthBig, **kwargs),      # type: ignore[misc]
        MultiLine(*args, color=colorSmall, width=widthSmall, **kwargs),  # type: ignore[misc]
    )


# # # Wrapper Elements
def Line(p1: XY, p2: XY, **kwargs) -> MultiLine:
    '''Draw a single line segment.

    Args:
        p1: First line segment vertex.
        p2: Second line segment vertex.
        kwargs: Passed to `MultiLine`.
    '''
    return MultiLine((p1, p2), **kwargs)


def RectangleCenter(c: XY, size_x: int, size_y: int, **kwargs) -> Rectangle:
    '''Draw a simple rectangle, from a center point and its size.

    Note that since integer coordinates are used, this might be off from
    the "real" center by at most half pixel.

    Args:
        c: Center point.
        size_x: Size of rectangle, side parallel to X axis.
        size_y: Size of rectangle, side parallel to Y axis.
        kwargs: Passed to `Rectangle`.
    '''
    delta = Vector(size_x // 2, size_y // 2)
    return Rectangle(c - delta, c + delta, **kwargs)


def SquareCenter(c: XY, size: int, *args, **kwargs) -> Rectangle:
    '''Draw a simple square, from a center point and its size.

    Note that since integer coordinates are used, this might be off from
    the "real" center by at most half pixel.

    Args:
        c: Center point.
        size: Side of the square.
        kwargs: Passed to `Rectangle`.
    '''
    return RectangleCenter(c, size, size, *args, **kwargs)


def EllipseCenter(c: XY, a: int, b: int, **kwargs) -> Ellipse:
    '''Draw a simple ellipse, from a center point and its axes.

    Args:
        c: Center point.
        a: Semi-Axis parallel to X.
        b: Semi-Axis parallel to Y.
        kwargs: Passed to `Ellipse`.
    '''
    axes_vector = Vector(a, b)
    return Ellipse(c - axes_vector, c + axes_vector, **kwargs)


def CircleCenter(c: XY, radius: int, *args, **kwargs) -> Ellipse:
    '''Draw a circle, from a center point and its radius.

    Note that since integer coordinates are used, this might be off from
    the "real" radius by at most half pixel.

    Args:
        c: Center point
        radius: Circle radius.
        kwargs: Passed to `Ellipse`.
    '''
    assert radius == int(radius)
    return EllipseCenter(c, a=radius, b=radius, **kwargs)



'''
All custom exceptions raised in the project.
'''
import typing

if typing.TYPE_CHECKING:
    from .diagram import DiagramElement


class InvalidWidgetDefinition(ValueError):
    pass


class InvalidWidgetState(Exception):
    def __init__(self, msg: str):
        self.msg = msg


class InvalidLayoutError(Exception):
    pass


class InvalidCallbackDefinition(Exception):
    def __init__(self, msg: str):
        self.msg = msg


class InvalidImageKey(Exception):
    def __init__(self, key: str):
        self.key = key


class InvalidImageType(Exception):
    def __init__(self, dtype: str):
        self.dtype = dtype


class InvalidRender(Exception):
    def __init__(self, de: 'DiagramElement'):
        self.de = de


class EventLoopError(Exception):
    def __init__(self, msg: str):
        self.msg = msg


class EventLoopConcurrencyError(EventLoopError):
    pass


class EventBusError(Exception):
    def __init__(self, msg: str):
        self.msg = msg


class EventBusRegisterError(EventBusError):
    pass



# import typing
import unittest
import logging

from tkmilan.validation import StaticList, StaticMap, StaticMapLabels, StaticMapValues
from tkmilan.validation import LimitBounded, LimitUnbounded
from tkmilan import fn

logger = logging.getLogger(__name__)


class Test_StaticList(unittest.TestCase):
    def test_valid(self):
        spec_basic = StaticList(['1', '2', '3'], default='2')
        with self.subTest(spec=spec_basic):
            self.assertTrue('2' in spec_basic)
            self.assertFalse('10' in spec_basic)
            self.assertEqual(len(spec_basic), 3)

    def test_errors(self):
        with self.assertRaises(ValueError):
            StaticList(['1', '2', '3'], default='5')
        with self.assertRaises(ValueError):
            StaticList(['1', '2', '3'], defaultIndex=5)


class Test_StaticMap(unittest.TestCase):
    map_basic = {'1': 1, '2': 2, '3': 3}

    def test_valid(self):
        spec_basic = StaticMap(self.map_basic, defaultLabel='2')
        with self.subTest(spec=spec_basic):
            self.assertTrue('2' in spec_basic)
            self.assertTrue(2 in spec_basic.rlabels)
            self.assertFalse('5' in spec_basic)
            self.assertFalse(5 in spec_basic.rlabels)
            self.assertEqual(len(spec_basic), 3)

    def test_valid_wrappers(self):
        spec_basic_labels = StaticMapLabels(int, tuple(self.map_basic.keys()), defaultIndex=0)
        with self.subTest(spec=spec_basic_labels):
            self.assertTrue('2' in spec_basic_labels)
            self.assertTrue(2 in spec_basic_labels.rlabels)
            self.assertFalse('5' in spec_basic_labels)
            self.assertFalse(5 in spec_basic_labels.rlabels)
            self.assertTrue('1', spec_basic_labels.ldefault)
        spec_basic_values = StaticMapValues(str, tuple(self.map_basic.values()), defaultIndex=-1)
        with self.subTest(spec=spec_basic_values):
            self.assertTrue('2' in spec_basic_values)
            self.assertTrue(2 in spec_basic_values.rlabels)
            self.assertFalse('5' in spec_basic_values)
            self.assertFalse(5 in spec_basic_values.rlabels)
            self.assertTrue('3', spec_basic_values.ldefault)

    def test_errors(self):
        with self.assertRaises(ValueError):
            StaticMap(self.map_basic, defaultLabel='5')
        with self.assertRaises(ValueError):
            StaticMap(self.map_basic, defaultValue=5)
        with self.assertRaises(AssertionError):
            StaticMapLabels(str, tuple(self.map_basic.values()), defaultIndex=0)


class Test_Limit(unittest.TestCase):
    def test_int_bounded(self):
        lim_imin = LimitBounded(1, 10, fn=fn.valNumber, imin=True, imax=False)
        with self.subTest(limit=str(lim_imin)):
            self.assertTrue('1' in lim_imin)
            self.assertTrue('5' in lim_imin)
            self.assertFalse('10' in lim_imin)
        lim_imax = LimitBounded('1', '10', fn=fn.valNumber, imin=False, imax=True)
        with self.subTest(limit=str(lim_imax)):
            self.assertFalse('1' in lim_imax)
            self.assertTrue('5' in lim_imax)
            self.assertTrue('10' in lim_imax)
        lim_inone = LimitBounded('1', '10', fn=fn.valNumber, imin=False, imax=False, default='5')
        with self.subTest(limit=str(lim_inone)):
            self.assertFalse('1' in lim_inone)
            self.assertTrue('5' in lim_inone)
            self.assertFalse('10' in lim_inone)

    def test_int_infinite(self):
        lim_nomin = LimitUnbounded(None, 10, fn=fn.valNumber)
        with self.subTest(limit=str(lim_nomin)):
            self.assertTrue('-100' in lim_nomin)
            self.assertTrue('5' in lim_nomin)
            self.assertFalse('+100' in lim_nomin)
        lim_nomax = LimitUnbounded('1', None, fn=fn.valNumber)
        with self.subTest(limit=str(lim_nomax)):
            self.assertFalse('-100' in lim_nomax)
            self.assertTrue('5' in lim_nomax)
            self.assertTrue('+100' in lim_nomax)

    def test_errors(self):
        for cls in (LimitBounded, LimitUnbounded):
            with self.assertRaises(ValueError):
                cls('1', '10', fn=fn.valNumber, imin=False, imax=False)  # Default Default 0 not in range
            with self.assertRaises(ValueError):
                # Weirdness: Strange `fn`
                cls('x', 'xxx', fn=len, default=2)  # No default roundtrip: str(2) == '2'; len('2') == 1; 1 != 2


if __name__ == '__main__':
    import sys
    logs_lvl = logging.DEBUG if '-v' in sys.argv else logging.INFO
    logging.basicConfig(level=logs_lvl, format='%(levelname)5.5s:%(funcName)s: %(message)s', stream=sys.stderr)
    unittest.main()



'''
Variable classes. Extends the existing variable types defined in `tkinter`.
'''
import typing
import tkinter as tk
from functools import wraps
import collections.abc

from . import model as tkmilan_model
from . import validation
if typing.TYPE_CHECKING:
    from . import mixin as tkmilan_mixin
    from . import RootWindow

Variable = tk.Variable

Boolean = tk.BooleanVar
'''Value holder for `bool`.'''
Double = tk.DoubleVar
'''Value holder for `float`.'''
Int = tk.IntVar
'''Value holder for `int`.'''
String = tk.StringVar
'''Value holder for `str`.'''

# TypeVar
objT = typing.TypeVar('objT')
'''Generic Type-Checking variable type for `ObjectList`.

This is a `typing.TypeVar`, to be used only when typechecking. See `typing`.
'''
spvT = typing.TypeVar('spvT')
'''Generic Type-Checking variable type for `SpecParsed`.

This is a `typing.TypeVar`, to be used only when typechecking. See `typing`.
'''


def trace(var: Variable, function: typing.Callable, *,
          trace_mode: tkmilan_model.TraceModeT = 'write',
          trace_initial: bool = False,
          **kwargs: typing.Any) -> str:
    '''Trace the variable ``var``.

    There is no Python documentation, see ``Tk`` :tk:`trace variable
    <trace.html#M14>` documentation.

    The ``function`` arguments will be:

      - ``var``: The traced variable
      - ``etype``: The trace mode (see ``trace_mode``)
      - The other given ``kwargs``

    It is possible to run the function as soon as the trace is defined. This is
    useful if the trace is used to keep state in sync.

    Args:
        var: The traced variable.
        function: The callback function.
        trace_mode: The trace mode, when should the callback be invoked.
            `None` when running on the initial function call (see ``trace_initial``).
        trace_initial: Run the callback function on start.
            Runs async (similar to `model.TimeoutIdle`).
        kwargs: Passed to the callback function.

    Returns:
        Return the function identifier, as stored by ``Tk``.
    '''
    assert not isinstance(var, nothing), f'{var}: Tracing "nothing" is unsupported'

    @wraps(function)
    def trace_wrapper(name: str, index: str, etype: str):  # "Real" tk function
        assert isinstance(name, str) and isinstance(etype, str)
        if __debug__:
            assert etype in ['read', 'write', 'unset'], f'Unsupported trace mode: {etype}'
            etype = typing.cast(tkmilan_model.TraceModeT, etype)
        return function(var, etype, **kwargs)

    if trace_initial:
        assert hasattr(var, '_root'), f'Invalid variable object: {var!r}'

        @wraps(function)
        def trace_wrapper_initial():
            return function(var, None, **kwargs)

        rw: 'RootWindow' = var._root
        rw.after_idle(trace_wrapper_initial)  # No need for a `model.TimeoutIdle` here
    return var.trace_add(
        trace_mode,
        trace_wrapper,
    )


# Specification Tower
class Spec(String):
    '''Abstract parent class for validated `String` variables.

    This variable tracks the label value, but also its valid states, which
    makes it so that the state type is a `model.VState`, not a simple `str`.

    Note that this might specify a complex validation, not just a list of
    discrete values, nor a continous range, for example.

    .. note::

        This should not be instanced directly, it's a parent class to be subclassed
        on other variable types.

    See Also:
        For a countable variant of this, see `SpecCountable`. For a valued
        variant of this, see `SpecParsed`.

    .. automethod:: __contains__
    '''
    container: typing.Container[str]
    '''The container object representing the specification.

    Follows the `container <collections.abc.Container>` class interface.
    '''
    ldefault: str
    '''The default label to be shown to the user.'''

    def __new__(cls,
                spec: typing.Optional[validation.VSpec] = None,
                *args, **kwargs):
        if cls in (Spec, SpecCountable, SpecParsed):
            if isinstance(spec, validation.VSpec):
                if __debug__:
                    spec_class = spec.__class__
                    new_class = MAP_VALIDATIONS[spec.__class__]
                assert spec_class in MAP_VALIDATIONS, f'{spec_class.__qualname__}: No mapping for specified variable'
                assert issubclass(new_class, cls), f'{new_class.__qualname__}: Not a subclass for {cls.__qualname__}'
                return MAP_VALIDATIONS[spec.__class__](spec, *args, **kwargs)
            else:
                raise ValueError(f'{cls.__qualname__}: Do not instance directly, abstract class')
        else:
            return super().__new__(cls)

    # @typing.override
    def get(self) -> tkmilan_model.VState[bool]:  # type: ignore[override]
        '''Returns the value of the variable, as `VState`.

        The ``value`` alternates between `True` for valid labels, and `None`
        for invalid labels.
        '''
        label: str = super().get()
        assert isinstance(self.container, collections.abc.Container), f'{self} missing container'
        # Use True/None for "automatic" validation
        value = True if label in self.container else None
        return tkmilan_model.VState(label=label, value=value)

    def set(self, state: typing.Union[str, tkmilan_model.VState[bool]]):
        '''Set the variable to ``state``.

        Only the label is set, either for direct `str` or `VState`.

        .. note::

            In debug mode, when setting the variable using a `VState`, the
            ``value`` itself is re-validated.
        '''
        assert isinstance(self.container, collections.abc.Container), f'{self} missing container'
        label: str  # Support setting labels and VState
        if isinstance(state, tkmilan_model.VState):
            label = state.label
            assert state.value is (True if label in self.container else None), f'Weird value:: {state}'
        else:  # Setting a simple label
            assert isinstance(state, str)
            label = state
        return super().set(label)

    def __contains__(self, label: str) -> bool:
        '''Check if the ``label`` label satisfies the specification.

        Wraps the `container` interface.

        Should be used as ``label in Spec``.
        '''
        assert isinstance(self.container, collections.abc.Container), f'{self} missing container'
        return label in self.container


class SpecCountable(Spec):
    '''Abstract parent class for validated `String` variables, with countable
    amount of valid alternatives.

    This is an alternative to `Spec`, when the amount of possible valid values is countable.

    .. note::

        This should not be instanced directly, it's a parent class to be subclassed
        on other variable types.

    See Also:
        For a simpler version of this, see `Spec`. For a valued variant of the
        base class, see `SpecParsed`.

    .. automethod:: __contains__
    .. automethod:: __len__
    '''
    def lall(self) -> typing.Sequence[str]:
        '''Get all valid labels supported by this variable.'''
        # Default Implementation
        assert isinstance(self.container, collections.abc.Sized), f'Variable "{self}" missing sizable container'
        assert isinstance(self.container, typing.Iterable), f'Variable "{self}" missing iterable container'
        return tuple(self.container)

    def __len__(self) -> int:
        '''Count how many labels this specification contains.

        Should be used as ``len(SpecCountable)``.
        '''
        assert isinstance(self.container, collections.abc.Sized), f'{self} missing sizable container'
        return len(self.container)


class SpecParsed(Spec, typing.Generic[spvT]):
    '''Abstract parent class for validated `String` variables, with a parsed
    value (instead of a boolean validation).

    .. note::

        This should not be instanced directly, it's a parent class to be subclassed
        on other variable types.

    See Also:
        For a simpler version of this, see `Spec`. For a countable variant of
        the base class, see `SpecCountable`.
    '''
    container: typing.Mapping[str, spvT]
    '''The container object representing the specification, a mapping between
    label and value.

    Follows the `mapping <collections.abc.Mapping>` class interface.
    '''
    lempty: str = ''
    '''The default label for invalid values.

    This is only necessary on rare circumstances.
    Should map to an invalid value.
    '''

    # @typing.override
    def get(self) -> tkmilan_model.VState[typing.Optional[spvT]]:  # type: ignore[override]
        '''Returns the value of the variable, as `VState`.

        The ``value`` alternates between the parsed label, and `None` for
        invalid labels.

        See Also:
            The base class version is `Spec.get`. It is compatible, but with
            different semantics.
        '''
        label: str = super(String, self).get()
        return tkmilan_model.VState(label=label, value=self.container.get(label))

    # @typing.override
    def set(self, state: typing.Union[str, tkmilan_model.VState[typing.Optional[spvT]]]):  # type: ignore[override]
        '''Set the variable to ``state``.

        Only the label is set, either for direct `str` or `VState`.

        .. note::

            In debug mode, when setting the variable using a `VState`, the
            ``value`` itself is re-validated.

        See Also:
            The base class version is `Spec.set`. It is compatible, but with
            different semantics.
        '''
        label: str  # Support setting labels and VState
        if isinstance(state, tkmilan_model.VState):
            label = state.label
            assert state.value == self.container.get(label), f'{self} has a weird value'
        else:  # Setting a simple label
            assert isinstance(state, str)
            label = state
        return super(String, self).set(label)

    def getLabel(self, value: typing.Optional[spvT]) -> str:
        '''Return the first label corresponding to the given ``value``.

        This is not very efficient, and the mapping might not be bi-injective.
        It's mostly implemented for completeness' sake.

        Returns:
            `lempty` if the value is not valid.
        '''
        # This is very inneficcient, but always works
        for label, cvalue in self.container.items():
            if value == cvalue:
                return label
        assert self.lempty not in self.container, f'{self} has an valid "empty" label: "{self.lempty}"'
        return self.lempty


# Other Special Variables
class nothing(Variable):
    '''Value holder for `None`.

    Useful for widgets that don't store anything, like buttons.
    '''
    def get(self):
        return None

    def set(self, value: None):
        pass


class aggregator(Variable):
    '''Synthetic value holder for an aggregation of other variables.

    Useful for container widgets.
    '''
    _default: str = ''
    tout: typing.Optional[tkmilan_model.TimeoutIdle]
    '''Hold the `TimeoutIdle <model.TimeoutIdle>` object that indicates
    this variable is setup.

    Since this variable is a synthetic variable, useful only when triggered
    from other function, it needs configuration before it can be used. For
    optimisation reasons, this is not done right on ``__init__`` like most
    other variables.

    Defaults to `None`, indicating it is not ready. See `ready`.
    '''
    # TODO: Save the child variables? `children: typing.Set[Variable]`
    def __init__(self, master=None, value=None, name=None, *, cwidget: 'tkmilan_mixin.MixinWidget'):
        super().__init__(master, value, name)
        self.tout = None
        self.cwidget = cwidget

    @property
    def ready(self) -> bool:
        '''Check if the variable is ready for usage.'''
        return self.tout is not None

    def get(self) -> None:
        assert self.ready, f'{self}: Unprepared aggregator variable'
        super().get()  # Trigger variable read
        return None

    def set(self, value: None = None):
        assert self.ready, f'{self}: Unprepared aggregator variable'
        return super().set('')  # Trigger variable write


# Actual Implementations
class StringList(tk.Variable):
    '''Value holder for a list of `str`.

    In ``Tk``, everything is a string and the syntax for lists is similar to
    Python, so this is technically supported, but sounds like a coincidence.

    Works well, though.

    See Also:
        `ObjectList` for an arbitrary list of Python objects.
    '''
    _default: typing.Iterable = []

    def get(self) -> typing.Iterable[str]:
        return [x for x in super().get()]

    def set(self, value: typing.Iterable[str]) -> None:
        return super().set([x for x in value])


class ObjectList(tk.Variable, typing.Generic[objT]):
    '''Generic value holder for a sequence of object of `objT` type.

    Just keep an instance variable with the "actual" value. Pretend the value
    is just an empty `str`.

    See Also:
        `StringList`: Simpler version of this, supports only `str`.
    '''
    # The dummy read/writes are necessary for traces to work correctly
    _default: typing.Sequence[objT] = []
    __actual_value: typing.Optional[typing.Sequence[objT]] = None

    def get(self) -> typing.Sequence[objT]:
        super().get()  # Dummy read
        return self.__actual_value or list(self._default)

    def set(self, value: typing.Sequence[objT]) -> None:
        self.__actual_value = value
        return super().set('')  # Dummy write


class StaticList(SpecCountable):
    '''Specifies a static list of labels.

    Args:
        spec: The list validation object.
        name: Name of the variable (passed to `String`).
            Optional, defaults to an autogenerated name.
    '''
    def __init__(self, spec: validation.StaticList, *,
                 name: typing.Optional[str] = None,
                 # Do not document "master", internal-ish
                 master: typing.Optional[tk.Widget] = None,
                 ):
        assert isinstance(spec, validation.StaticList), f'Invalid Spec: {spec}'
        super().__init__(master=master, name=name, value=spec.ldefault)
        # Set `Spec` parameters
        self.container = spec
        self.ldefault = spec.ldefault


class StaticMap(SpecCountable, SpecParsed[spvT]):
    '''Specifies a static mapping between labels and its corresponding values.

    Args:
        spec: The mapping validation object.
        name: Name of the variable (passed to `String`).
            Optional, defaults to an autogenerated name.
    '''
    def __init__(self, spec: validation.StaticMap[spvT], *,
                 name: typing.Optional[str] = None,
                 # Do not document "master", internal-ish
                 master: typing.Optional[tk.Widget] = None,
                 ):
        assert isinstance(spec, validation.StaticMap), f'Invalid Spec: {spec}'
        super().__init__(master=master, name=name, value=spec.ldefault)
        assert self.lempty not in spec, f'{self.__class__.__qualname__}: Valid Empty String: {self.lempty}'
        # Set `SpecCountable`/`SpecParsed` parameters
        self.container = spec
        self.rlabels = spec.rlabels
        self.ldefault = spec.ldefault

    def getLabel(self, value: typing.Optional[spvT]) -> str:
        '''Return the label corresponding to the given ``value``.

        This is very efficient, and the mapping is enforced to be bi-injective.

        See Also:
            The upstream function this aliases is `SpecParsed.getLabel`.
        '''
        if value is None:
            return self.lempty
        else:
            return self.rlabels.get(value, self.lempty)


class Limit(SpecParsed[int]):
    '''Represent a range to limit a value, for simple integers.

    The range is represented as the ``container`` parameter, one of the
    `LimitBounded <validation.LimitBounded>` or ``LimitUnbounded
    <validation.LimitUnbounded>``.

    When instancing directly, the ``limit`` parameter will instance the right
    subclass. This means this class can not be instanced directly without
    arguments, and when instance with a validation representation, it will
    create an object of the right subclass.

    Arguments:
        limit: The validation representation for this type. **Mandatory**.

    All other arguments are passed to the right subclass.

    See Also:
        The supported subclasses:

        - `LimitBounded`
        - `LimitUnbounded`
    '''
    container: typing.Union[validation.LimitBounded, validation.LimitUnbounded]

    def __new__(cls,
                limit: typing.Union[validation.LimitBounded, validation.LimitUnbounded],
                *args, **kwargs,
                ):
        if cls is Limit:
            if isinstance(limit, validation.LimitBounded):
                return LimitBounded(limit, *args, **kwargs)
            elif isinstance(limit, validation.LimitUnbounded):
                return LimitUnbounded(limit, *args, **kwargs)
            else:
                raise NotImplementedError
        else:
            # Sub-Class, instance as usual
            return super().__new__(cls)

    def get_spinargs(self) -> typing.Tuple[float, float, float]:
        '''Get all `SpinboxN` value arguments.

        Return Value:
            Returns a tuple with three floating point values:

            - ``from``
            - ``to``
            - ``increment``
        '''
        infinite: float = float('+inf')
        sincrement = 1.0  # TODO: Support "step"
        sfrom: float
        if self.container.min is None:
            sfrom = -1 * infinite
        else:
            assert isinstance(self.container.min, typing.SupportsFloat), f'Invalid min value: {self.container.min!r}'
            sfrom = float(self.container.min)
        sto: float
        if self.container.max is None:
            sto = +1 * infinite
        else:
            assert isinstance(self.container.max, typing.SupportsFloat), f'Invalid max value: {self.container.max!r}'
            sto = float(self.container.max)
        if not self.container.imin:
            sfrom += sincrement
        if not self.container.imax:
            sto -= sincrement
        return sfrom, sto, sincrement

    def getLabel(self, value: typing.Optional[spvT]) -> str:
        '''Return the label corresponding to the given ``value``.

        This is very efficient.

        See Also:
            The upstream function this aliases is `SpecParsed.getLabel`.
        '''
        return self.container.get_label(value, invalid=self.lempty)


class LimitBounded(Limit, SpecCountable):
    '''A bounded variant of the `Limit` range.

    Since this is bounded, it's a countable specification.

    Args:
        limit: The `LimitBounded <validation.LimitBounded>` range object..
        name: Name of the variable (passed to `String`).
            Optional, defaults to an autogenerated name.
    '''
    def __init__(self, limit: validation.LimitBounded, *,
                 name: typing.Optional[str] = None,
                 # Do not document "master", internal-ish
                 master: typing.Optional[tk.Widget] = None,
                 ):
        assert isinstance(limit, validation.LimitBounded), f'Invalid LimitBounded: {limit}'
        super().__init__(master=master, name=name, value=limit.ldefault)
        # Set `SpecCountable`/`SpecParsed` parameters
        self.container: validation.LimitBounded = limit
        self.ldefault = limit.ldefault


class LimitUnbounded(Limit):
    '''An unbounded variant of the `Limit` range.

    Since this is unbounded, it's not a countable specification.

    Args:
        limit: The `LimitUnbounded <validation.LimitUnbounded>` range object.
        name: Name of the variable (passed to `String`).
            Optional, defaults to an autogenerated name.
    '''
    def __init__(self, limit: validation.LimitUnbounded, *,
                 name: typing.Optional[str] = None,
                 # Do not document "master", internal-ish
                 master: typing.Optional[tk.Widget] = None,
                 ):
        assert isinstance(limit, validation.LimitUnbounded), f'Invalid LimitUnbounded: {limit}'
        super().__init__(master=master, name=name, value=limit.ldefault)
        # Set `SpecCountable`/`SpecParsed` parameters
        self.container: validation.LimitUnbounded = limit
        self.ldefault = limit.ldefault


MAP_VALIDATIONS = {
    validation.StaticList: StaticList,
    validation.StaticMap: StaticMap,
    validation.LimitBounded: LimitBounded,
    validation.LimitUnbounded: LimitUnbounded,
}
'''Mapping between `validation.VSpec` representations, and the corresponding variable class.'''
if __debug__:
    vspec_subclasses = set(validation.VSpec.__subclasses__())
    map_subclasses = set(MAP_VALIDATIONS.keys())
    delta_subclasses_names = [cls.__qualname__ for cls in map_subclasses.symmetric_difference(vspec_subclasses)]
    assert len(delta_subclasses_names) == 0, f'Missing `MAP_VALIDATIONS` for {" ".join(delta_subclasses_names)}'



# import typing
import unittest
import logging

from tkmilan import diagram as d

logger = logging.getLogger(__name__)


class Test_Diagram_Geometry(unittest.TestCase):
    def test_line(self):
        ln_1 = d.GeometryLine.points(d.XY(0, 0), d.XY(10, 10))
        logger.debug('ln_1: %s', ln_1.equation)
        self.assertEqual(ln_1, d.GeometryLine(1.0, 0.0))
        ln_2 = d.GeometryLine.points(d.XY(0, 0), d.XY(1, -1))
        logger.debug('ln_2: %s', ln_2.equation)
        self.assertEqual(ln_2, d.GeometryLine(-1.0, 0.0))
        # Intersections
        self.assertEqual(ln_1.intersect_line(ln_2), d.XY(0, 0))

    def test_line_intersect(self):
        ln = d.GeometryLine(1, 0)
        logger.debug('ln: %s', ln.equation)
        self.assertEqual(ln.intersect_x(0), d.XY(0, 0))
        self.assertEqual(ln.intersect_x(10), d.XY(10, 10))
        self.assertEqual(ln.intersect_y(10), d.XY(10, 10))
        ln_h = d.GeometryLine(0, 10)
        logger.debug('ln_h: %s', ln_h.equation)
        self.assertEqual(ln_h.intersect_x(10), d.XY(10, 10))
        self.assertIsNone(ln_h.intersect_y(0))   # Parallel
        self.assertIsNone(ln_h.intersect_y(10))  # Coincident
        ln_v = d.GeometryLine(None, 10)
        logger.debug('ln_v: %s', ln_v.equation)
        self.assertIsNone(ln_v.intersect_x(0))   # Parallel
        self.assertIsNone(ln_v.intersect_x(10))  # Coincident
        self.assertEqual(ln_v.intersect_y(10), d.XY(10, 10))
        # Other
        self.assertEqual(ln.intersect_x(ln_v.c0), ln.intersect_line(ln_v))
        self.assertEqual(ln.intersect_y(ln_h.c0), ln.intersect_line(ln_h))

    def test_line_parallel(self):
        ln = d.GeometryLine(2, 0)
        p_i, p_n = d.XY(2, 4), d.XY(4, 2)
        self.assertIn(p_i, ln)  # Point in Line
        self.assertNotIn(p_n, ln)
        p_1 = d.XY(0, 2)
        ln_1 = ln.parallel_point(p_1)
        self.assertEqual(ln_1, d.GeometryLine(2, 2))
        p_2 = d.XY(1, 0)
        ln_2 = ln.parallel_point(p_2)
        self.assertEqual(ln_2, d.GeometryLine(2, -2.0))
        # Line by Point and Slope
        self.assertEqual(ln_2, d.GeometryLine.point_slope(p_2, ln.m))

    def test_line_perpendicular(self):
        ln_1 = d.GeometryLine(2, 20)
        ln_1p = d.GeometryLine(-1.0 / 2, 0)
        pi_1 = ln_1.intersect_line(ln_1p)
        po_1 = d.XY(10, 5)
        ln_1o = d.GeometryLine(ln_1p.m, po_1.y - ln_1p.m * po_1.x)
        self.assertIn(pi_1, ln_1)
        self.assertEqual(ln_1.perpendicular_point(pi_1), ln_1p)
        self.assertEqual(ln_1.perpendicular_point(po_1), ln_1o)
        self.assertEqual(ln_1p.parallel_point(po_1), ln_1o)


if __name__ == '__main__':
    import sys
    logs_lvl = logging.DEBUG if '-v' in sys.argv else logging.INFO
    logging.basicConfig(level=logs_lvl, format='%(levelname)5.5s:%(funcName)s: %(message)s', stream=sys.stderr)
    unittest.main()
